import numpy as np

from ..config import create_config
from ..data import data_shape, split_generator
from ..tensorflow_wrapper import tf
from .model import Model, clip_log, sum_gradient, sum_hessian
from .opt_int import build_amp, sum_gradient_data2

set_function, get_function, register_function = create_config()


@register_function("default_bg")
def f_bg(data):
    return data.get("bg_value", tf.ones((data_shape(data),), dtype="float64"))


@register_function("default_eff")
def f_eff(data):
    return data.get("eff_value", tf.ones((data_shape(data),), dtype="float64"))


class Model_cfit(Model):
    def __init__(self, amp, w_bkg=0.001, bg_f=None, eff_f=None):
        super().__init__(amp, w_bkg)
        if bg_f is None:
            bg_f = get_function("default_bg")
        elif isinstance(bg_f, str):
            bg_f = get_function(bg_f)
        if eff_f is None:
            eff_f = get_function("default_eff")
        elif isinstance(eff_f, str):
            eff_f = get_function(eff_f)
        self.bg = bg_f
        self.vm = amp.vm
        self.w_bkg = w_bkg
        self.eff = eff_f
        self.sig = lambda x: self.eff(x) * self.Amp(x)

    def nll(
        self,
        data,
        mcdata,
        weight: tf.Tensor = 1.0,
        batch=None,
        bg=None,
        mc_weight=None,
    ):
        """
        Calculate NLL.

        .. math::
          -\\ln L = -\\sum_{x_i \\in data } w_i \\ln P(x_i;\\theta_k) \\\\
          P(x_i;\\theta_k) = (1-f_{bg})Amp(x_i; \\theta_k) + f_{bg} Bg(x_{i};\\theta_k)

        :param data: Data array
        :param mcdata: MCdata array
        :param weight: Weight of data???
        :param batch: The length of array to calculate as a vector at a time. How to fold the data array may depend on the GPU computability.
        :param bg: Background data array. It can be set to ``None`` if there is no such thing.
        :return: Real number. The value of NLL.
        """
        data, weight = self.get_weight_data(data, weight)
        sw = tf.reduce_sum(weight)
        sig_data = self.sig(data)
        bg_data = self.bg(data)
        if mc_weight is None:
            int_mc = tf.reduce_mean(self.sig(mcdata))
            int_bg = tf.reduce_mean(self.bg(mcdata))
        else:
            int_mc = tf.reduce_sum(mc_weight * self.sig(mcdata))
            int_bg = tf.reduce_sum(mc_weight * self.bg(mcdata))
        ln_data = tf.math.log(
            (1 - self.w_bkg) * sig_data / int_mc
            + self.w_bkg * bg_data / int_bg
        )
        nll_0 = -tf.reduce_sum(tf.cast(weight, ln_data.dtype) * ln_data)
        return nll_0

    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        r"""
        .. math::
            P = (1-frac) \frac{amp(data)}{\sum amp(phsp)} + frac \frac{bg(data)}{\sum bg(phsp)}

        .. math::
            nll = - \sum log(p)

        .. math::
           \frac{\partial nll}{\partial \theta} = -\sum \frac{1}{p} \frac{\partial p}{\partial \theta}
           = - \sum \frac{\partial \ln \bar{p}}{\partial \theta }  +\frac{\partial nll}{\partial I_{sig} }
           \frac{\partial I_{sig} }{\partial \theta}
           +\frac{\partial nll}{\partial I_{bg}}\frac{\partial I_{sig}}{\partial \theta}

        """
        var = self.vm.trainable_variables
        mcdata = list(mcdata)
        mc_weight = list(mc_weight)
        int_sig, g_int_sig = sum_gradient(self.sig, mcdata, var, mc_weight)
        int_bg, g_int_bg = sum_gradient(self.bg, mcdata, var, mc_weight)
        v_int_sig, v_int_bg = (
            tf.Variable(int_sig, dtype="float64"),
            tf.Variable(int_bg, dtype="float64"),
        )

        def prob(x):
            return (1 - self.w_bkg) * self.sig(
                x
            ) / v_int_sig + self.w_bkg * self.bg(x) / v_int_bg

        ll, g_ll = sum_gradient(
            prob, data, var + [v_int_sig, v_int_bg], weight, trans=clip_log
        )
        g_ll_sig, g_ll_bg = g_ll[-2], g_ll[-1]
        g = [
            -g_ll[i] - g_int_sig[i] * g_ll_sig - g_int_bg[i] * g_ll_bg
            for i in range(len(var))
        ]
        return -ll, g

    def nll_grad_hessian(
        self, data, mcdata, weight=1.0, batch=24000, bg=None, mc_weight=1.0
    ):
        r"""
        The parameters are the same with ``self.nll()``, but it will return Hessian as well.

        .. math::
            \frac{\partial^2 f}{\partial x_i \partial x_j} =
            \frac{\partial y_k}{\partial x_i}
            \frac{\partial^2 f}{\partial y_k \partial y_l}
            \frac{\partial y_l}{\partial x_j}
            + \frac{\partial f}{\partial y_k} \frac{\partial^2 y_k}{\partial x_i \partial x_j}

        .. math::
            y = \{x_i; I_{sig}, I_{bg}\}

        .. math::
            \frac{\partial y_k }{\partial x_i} = (\delta_{ik};\frac{\partial I_{sig}}{\partial x_i}, \frac{\partial I_{bg}}{\partial x_i})

        :return NLL: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        :return Hessian: 2-D Array of real numbers. The Hessian matrix of the variables.
        """
        data, weight = self.get_weight_data(data, weight, bg=bg)
        if isinstance(mc_weight, float):
            mc_weight = tf.convert_to_tensor(
                [mc_weight] * data_shape(mcdata), dtype="float64"
            )
        n_mc = tf.reduce_sum(mc_weight)
        sw = tf.reduce_sum(weight)
        var = self.vm.trainable_variables
        int_sig, g_int_sig, h_int_sig = sum_hessian(
            self.sig,
            split_generator(mcdata, batch),
            var,
            weight=split_generator(mc_weight, batch),
        )

        int_bg, g_int_bg, h_int_bg = sum_hessian(
            self.bg,
            split_generator(mcdata, batch),
            var,
            weight=split_generator(mc_weight, batch),
        )

        v_int_sig, v_int_bg = (
            tf.Variable(int_sig, dtype="float64"),
            tf.Variable(int_bg, dtype="float64"),
        )

        def prob(x):
            return (1 - self.w_bkg) * self.sig(
                x
            ) / v_int_sig + self.w_bkg * self.bg(x) / v_int_bg

        ll, g_ll, h_ll = sum_hessian(
            prob,
            split_generator(data, batch),
            var + [v_int_sig, v_int_bg],
            weight=split_generator(weight, batch),
            trans=clip_log,
        )

        n_var = len(var)
        g_ll_sig, g_ll_bg = g_ll[-2], g_ll[-1]
        g = [
            -g_ll[i] - g_int_sig[i] * g_ll_sig - g_int_bg[i] * g_ll_bg
            for i in range(len(var))
        ]
        jac = np.concatenate([np.eye(n_var), [g_int_sig, g_int_bg]], axis=0)
        h = (
            np.dot(jac.T, np.dot(h_ll, jac))
            + g_ll_sig * h_int_sig
            + g_ll_bg * h_int_bg
        )
        return -ll, g, -h


class Model_cfit_cached(Model_cfit):
    def __init__(self, amp, w_bkg=0.001, bg_f=None, eff_f=None):
        super().__init__(amp, w_bkg, bg_f, eff_f)
        self.cached_amp = build_amp.build_amp2s(amp.decay_group)
        self.cached_data = {}

    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        var = self.vm.trainable_variables
        data_id = id(data)
        mc_id = id(mcdata)
        mcdata = list(mcdata)
        mc_weight = list(mc_weight)

        if data_id not in self.cached_data:
            self.cached_data[data_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in data
            ]

        if mc_id not in self.cached_data:
            self.cached_data[mc_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in mcdata
            ]

        int_sig, g_int_sig = sum_gradient_data2(
            self.cached_amp,
            self.Amp.trainable_variables,
            mcdata,
            self.cached_data[mc_id],
            weight=mc_weight,
        )

        int_bg, g_int_bg = sum_gradient(self.bg, mcdata, var, mc_weight)
        v_int_sig, v_int_bg = (
            tf.Variable(int_sig, dtype="float64"),
            tf.Variable(int_bg, dtype="float64"),
        )

        def prob(x, c_data):
            return (1 - self.w_bkg) * self.eff(x) * self.cached_amp(
                x, c_data
            ) / v_int_sig + self.w_bkg * self.bg(x) / v_int_bg

        ll, g_ll = sum_gradient_data2(
            prob,
            var + [v_int_sig, v_int_bg],
            data,
            self.cached_data[data_id],
            weight=weight,
            trans=clip_log,
        )
        g_ll_sig, g_ll_bg = g_ll[-2], g_ll[-1]
        g = [
            -g_ll[i] - g_int_sig[i] * g_ll_sig - g_int_bg[i] * g_ll_bg
            for i in range(len(var))
        ]
        return -ll, g



import functools

import numpy as np
import tensorflow as tf

from tf_pwa.data import data_split


def nll_grad(f, var, args=(), kwargs=None, options=None):
    kwargs = kwargs if kwargs is not None else {}
    options = options if options is not None else {}

    @functools.wraps(f)
    def f_w():
        with tf.GradientTape() as tape:
            ret = f(*args, **kwargs)
        g = tape.gradient(ret, var, unconnected_gradients="zero", **options)
        return ret, g

    return f_w


def cal_fitfractions(amp, mcdata, res=None, batch=None, args=(), kwargs=None):
    r"""
    defination:

    .. math::
        FF_{i} = \frac{\int |A_i|^2 d\Omega }{ \int |\sum_{i}A_i|^2 d\Omega }
        \approx \frac{\sum |A_i|^2 }{\sum|\sum_{i} A_{i}|^2}

    interference fitfraction:

    .. math::
        FF_{i,j} = \frac{\int 2Re(A_i A_j*) d\Omega }{ \int |\sum_{i}A_i|^2 d\Omega }
        = \frac{\int |A_i +A_j|^2  d\Omega }{ \int |\sum_{i}A_i|^2 d\Omega } - FF_{i} - FF_{j}

    gradients (for error transfer):

    .. math::
        \frac{\partial }{\partial \theta_i }\frac{f(\theta_i)}{g(\theta_i)} =
        \frac{\partial f(\theta_i)}{\partial \theta_i} \frac{1}{g(\theta_i)} -
        \frac{\partial g(\theta_i)}{\partial \theta_i} \frac{f(\theta_i)}{g^2(\theta_i)}

    """
    kwargs = kwargs if kwargs is not None else {}
    var = amp.trainable_variables
    # allvar = [i.name for i in var]
    cahced_res = amp.used_res
    if res is None:
        res = list(amp.res)
    n_res = len(res)
    fitFrac = {}
    err_fitFrac = {}
    g_fitFrac = [None] * n_res
    amp.set_used_res(res)
    weight = 1.0
    if batch is not None:
        weight = mcdata.get("weight", 1.0)
        mcdata = list(data_split(mcdata, batch))
        if not isinstance(weight, float):
            weight = list(data_split(weight, batch))
    int_mc, g_int_mc = sum_gradient(
        amp, mcdata, var=var, weight=weight, args=args, kwargs=kwargs
    )
    for i in range(n_res):
        for j in range(i, -1, -1):
            amp_tmp = amp
            if i == j:
                name = "{}".format(res[i])
                amp_tmp.set_used_res([res[i]])
            else:
                name = (str(res[i]), str(res[j]))
                amp_tmp.set_used_res([res[i], res[j]])
            int_tmp, g_int_tmp = sum_gradient(
                amp_tmp,
                mcdata,
                var=var,
                weight=weight,
                args=args,
                kwargs=kwargs,
            )
            if i == j:
                fitFrac[name] = int_tmp / int_mc
                gij = (
                    g_int_tmp / int_mc - (int_tmp / int_mc) * g_int_mc / int_mc
                )
                g_fitFrac[i] = gij
            else:
                fitFrac[name] = (
                    (int_tmp / int_mc)
                    - fitFrac["{}".format(res[i])]
                    - fitFrac["{}".format(res[j])]
                )
                gij = (
                    g_int_tmp / int_mc
                    - (int_tmp / int_mc) * g_int_mc / int_mc
                    - g_fitFrac[i]
                    - g_fitFrac[j]
                )
            # print(name,gij.tolist())
            err_fitFrac[name] = gij
    amp.set_used_res(cahced_res)
    return fitFrac, err_fitFrac


def cal_fitfractions_no_grad(
    amp, mcdata, res=None, batch=None, args=(), kwargs=None
):
    r"""
    calculate fit fractions without gradients.
    """
    kwargs = kwargs if kwargs is not None else {}
    var = amp.trainable_variables
    # allvar = [i.name for i in var]
    cahced_res = amp.used_res
    if res is None:
        res = list(amp.res)
    n_res = len(res)
    fitFrac = {}
    amp.set_used_res(res)
    weight = 1.0
    if batch is not None:
        weight = mcdata.get("weight", 1.0)
        mcdata = list(data_split(mcdata, batch))
        if not isinstance(weight, float):
            weight = list(data_split(weight, batch))
    int_mc = sum_no_gradient(
        amp, mcdata, var=var, weight=weight, args=args, kwargs=kwargs
    )
    for i in range(n_res):
        for j in range(i, -1, -1):
            amp_tmp = amp
            if i == j:
                name = "{}".format(res[i])
                amp_tmp.set_used_res([res[i]])
            else:
                name = "{}x{}".format(res[i], res[j])
                amp_tmp.set_used_res([res[i], res[j]])
            int_tmp = sum_no_gradient(
                amp_tmp,
                mcdata,
                var=var,
                weight=weight,
                args=args,
                kwargs=kwargs,
            )
            if i == j:
                fitFrac[name] = int_tmp / int_mc
            else:
                fitFrac[name] = (
                    (int_tmp / int_mc)
                    - fitFrac["{}".format(res[i])]
                    - fitFrac["{}".format(res[j])]
                )
    amp.set_used_res(cahced_res)
    return fitFrac


def sum_gradient(
    amp,
    data,
    var,
    weight=1.0,
    func=lambda x: x,
    grad=True,
    args=(),
    kwargs=None,
):
    kwargs = kwargs if kwargs is not None else {}
    # n_variables = len(var)
    if isinstance(weight, float):  # 给data乘weight
        weight = [weight] * len(data)
    nll_list = []
    g_list = []

    def f(d, w):
        amp2s = amp(d, *args, **kwargs)  # amp是振幅表达式
        l_a = func(amp2s)  # ampPDF转换成每组数据的NLL表达式
        return tf.reduce_sum(tf.cast(w, l_a.dtype) * l_a)

    for d, w in zip(data, weight):
        if grad:  # 是否提供grad表达式
            p_nll, a = nll_grad(
                f, var, args=(d, w)
            )()  # 调用上面定义的nll_grad，返回f(d,w)和f对var的导数
            g_list.append([i.numpy() for i in a])
        else:
            p_nll = f(d, w)
        nll_list.append(p_nll.numpy())
    nll = sum(nll_list)
    if grad:
        g = np.array(g_list).sum(0)
        return nll, g  # nll值和各var的导数g值
    return nll


sum_no_gradient = functools.partial(sum_gradient, grad=False)



import json
from pprint import pprint

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.experimental import extra_amp, extra_data
from tf_pwa.main import regist_subcommand
from tf_pwa.utils import error_print


def json_print(dic):
    """print parameters as json"""
    s = json.dumps(dic, indent=2)
    print(s, flush=True)


@regist_subcommand(name="fit")
def fit(config="config.yml", init_params="init_params.json", method="BFGS"):
    """
    simple fit script
    """
    # load config.yml
    config = ConfigLoader(config)

    # set initial parameters if have
    try:
        config.set_params(init_params)
        print("using {}".format(init_params))
    except Exception as e:
        if str(e) != "[Errno 2] No such file or directory: 'init_params.json'":
            print(e)
        print("\nusing RANDOM parameters", flush=True)

    # print("\n########### initial parameters")
    # json_print(config.get_params())

    # fit
    data, phsp, bg, inmc = config.get_all_data()
    try:
        fit_result = config.fit(batch=65000, method=method)
    except KeyboardInterrupt:
        config.save_params("break_params.json")
        raise
    except Exception as e:
        print(e)
        config.save_params("break_params.json")
        raise
    json_print(fit_result.params)
    fit_result.save_as("final_params.json")

    # calculate parameters error
    fit_error = config.get_params_error(fit_result, batch=13000)
    fit_result.set_error(fit_error)
    fit_result.save_as("final_params.json")
    pprint(fit_error)

    print("\n########## fit results:")
    for k, v in config.get_params().items():
        print(k, error_print(v, fit_error.get(k, None)))

    # plot partial wave distribution
    config.plot_partial_wave(fit_result, plot_pull=True)

    # calculate fit fractions
    phsp_noeff = config.get_phsp_noeff()
    fit_frac, err_frac = config.cal_fitfractions({}, phsp_noeff)

    print("########## fit fractions")
    fit_frac_string = ""
    for i in fit_frac:
        if isinstance(i, tuple):
            name = "{}x{}".format(*i)
        else:
            name = i
        fit_frac_string += "{} {}\n".format(
            name, error_print(fit_frac[i], err_frac.get(i, None))
        )
    print(fit_frac_string)



"""
This module provides functions to calculate the Wigner D-function.

:math:`D^{j}_{m_1,m_2}(\\alpha,\\beta,\\gamma)=e^{-im_1\\alpha}d^{j}_{m_1,m_2}(\\beta)e^{-im_2\\gamma}`,
where the expression of the Wigner d-function is

.. math:: d^{j}_{m_1,m_2}(\\beta) = \\sum_{l=0}^{2j} w_{l}^{(j,m_1,m_2)}\\sin^{l}(\\frac{\\beta}{2}) \\cos^{2j-l}(\\frac{\\beta}{2}),

where the weight :math:`w_{l}^{(j,m_1,m_2)}` in each term satisfies

.. math:: w^{(j,m_1,m_2)}_{l} = (-1)^{m_1-m_2+k}\\frac{\\sqrt{(j+m_1)!(j-m_1)!(j+m_2)!(j-m_2)!}}{(j-m_1-k)!(j+m_2-k)!(m_1-m_2+k)!k!}

when :math:`k=\\frac{l+m_2-m_1}{2} \\in [\\max(0,m_2-m_1),\\min(j-m_1,j+m_2)]`, and otherwise
:math:`w^{(j,m_1,m_2)}_{l} = 0`.

"""

import functools
import math

import numpy as np

from .tensorflow_wrapper import tf


def _spin_int(x):
    if isinstance(x, int):
        return x
    return int(x + 0.1)


@functools.lru_cache()
def _tuple_delta_D_trans(j, la, lb, lc):
    ln = _spin_int(2 * j + 1)
    s = np.zeros(shape=(ln, ln, len(la), len(lb), len(lc)))
    for i_a, la_i in enumerate(la):
        for i_b, lb_i in enumerate(lb):
            for i_c, lc_i in enumerate(lc):
                delta = lb_i - lc_i
                if abs(delta) <= j:
                    idx = la_i + j, delta + j, i_a, i_b, i_c
                    s[tuple(map(_spin_int, idx))] = 1.0
    return s


def delta_D_trans(j, la, lb, lc):
    """
    The decay from particle *a* to *b* and *c* requires :math:`|l_b-l_c|\\leqslant j`

    (ja,ja) -> (ja,jb,jc)???
    """
    la, lb, lc = map(tuple, (la, lb, lc))
    ret = _tuple_delta_D_trans(j, la, lb, lc)
    return ret


def delta_D_index(j, la, lb, lc):
    la, lb, lc = map(tuple, (la, lb, lc))
    ret = _tuple_delta_D_index(j, la, lb, lc)
    return ret


@functools.lru_cache()
def _tuple_delta_D_index(j, la, lb, lc):
    ln = _spin_int(2 * j + 1)
    ret = []
    max_idx = ln * ln
    for i_a, la_i in enumerate(la):
        for i_b, lb_i in enumerate(lb):
            for i_c, lc_i in enumerate(lc):
                delta = lb_i - lc_i
                if abs(delta) <= j:
                    ret.append(_spin_int((la_i + j) * ln + delta + j))
                else:
                    ret.append(max_idx)
    return ret


def Dfun_delta(d, ja, la, lb, lc=(0,)):
    """
    The decay from particle *a* to *b* and *c* requires :math:`|l_b-l_c|\\leqslant j`

    :math:`D_{ma,mb-mc} = \\delta[(m1,m2)->(ma, mb,mc))] D_{m1,m2}`
    """
    t = delta_D_trans(ja, la, lb, lc)
    ln = _spin_int(2 * ja + 1)
    t_trans = tf.reshape(t, (ln * ln, len(la) * len(lb) * len(lc)))

    t_cast = tf.cast(t_trans, d.dtype)
    # print(d[0])

    d = tf.reshape(d, (-1, ln * ln))

    ret = tf.matmul(d, t_cast)
    return tf.reshape(ret, (-1, len(la), len(lb), len(lc)))


def Dfun_delta_v2(d, ja, la, lb, lc=(0,)):
    """
    The decay from particle *a* to *b* and *c* requires :math:`|l_b-l_c|\\leqslant j`

    :math:`D_{ma,mb-mc} = \\delta[(m1,m2)->(ma, mb,mc))] D_{m1,m2}`
    """
    idx = delta_D_index(ja, la, lb, lc)
    ln = _spin_int(2 * ja + 1)
    # print(d[0])

    d = tf.reshape(d, (-1, ln * ln))
    over_d = tf.pad(d, [[0, 0], [0, 1]], mode="CONSTANT")
    # print(d.shape, over_d.shape)
    # zeros = tf.zeros((d.shape[0], 1), dtype=d.dtype)

    # over_d = tf.concat([d, zeros], axis=-1)
    ret = tf.gather(over_d, idx, axis=-1)
    return tf.reshape(ret, (-1, len(la), len(lb), len(lc)))


@functools.lru_cache()
def small_d_weight(j):  # the prefactor in the d-function of β
    """
    For a certain j, the weight coefficient with index (:math:`m_1,m_2,l`) is
    :math:`w^{(j,m_1,m_2)}_{l} = (-1)^{m_1-m_2+k}\\frac{\\sqrt{(j+m_1)!(j-m_1)!(j+m_2)!(j-m_2)!}}{(j-m_1-k)!(j+m_2-k)!(m_1-m_2+k)!k!}`,
    and :math:`l` is an integer ranging from 0 to :math:`2j`.

    :param j: Integer :math:`2j` in the formula???
    :return: Of the shape (**j** +1, **j** +1, **j** +1). The indices correspond to (:math:`l,m_1,m_2`)
    """
    ret = np.zeros(shape=(j + 1, j + 1, j + 1))

    def f(x):
        return math.factorial(x >> 1)

    for m in range(-j, j + 1, 2):
        for n in range(-j, j + 1, 2):
            for k in range(max(0, n - m), min(j - m, j + n) + 1, 2):
                l = (2 * k + (m - n)) // 2
                sign = (-1) ** ((k + m - n) // 2)
                tmp = sign * math.sqrt(
                    1.0 * f(j + m) * f(j - m) * f(j + n) * f(j - n)
                )
                tmp /= f(j - m - k) * f(j + n - k) * f(k + m - n) * f(k)
                ret[l][(m + j) // 2][(n + j) // 2] = tmp
    return ret


def small_d_matrix(theta, j):
    """
    The matrix element of :math:`d^{j}(\\theta)` is
    :math:`d^{j}_{m_1,m_2}(\\theta) = \\sum_{l=0}^{2j} w_{l}^{(j,m_1,m_2)}\\sin^{l}(\\frac{\\theta}{2}) \\cos^{2j-l}(\\frac{\\theta}{2})`

    :param theta: Array :math:`\\theta` in the formula
    :param j: Integer :math:`2j` in the formula???
    :return: The d-matrices array. Same length as theta
    """
    a = tf.reshape(tf.range(0, j + 1, 1), (1, -1))

    half_theta = 0.5 * theta

    sintheta = tf.reshape(tf.sin(half_theta), (-1, 1))
    costheta = tf.reshape(tf.cos(half_theta), (-1, 1))

    a = tf.cast(a, dtype=sintheta.dtype)
    s = tf.pow(sintheta, a)
    c = tf.pow(costheta, j - a)
    sc = s * c
    w = small_d_weight(j)

    w = tf.cast(w, sc.dtype)
    w = tf.reshape(w, (j + 1, (j + 1) * (j + 1)))
    ret = tf.matmul(sc, w)
    # ret = tf.einsum("il,lab->iab", sc, w)

    return tf.reshape(ret, (-1, j + 1, j + 1))


def exp_i(theta, mi):
    """
    :math:`e^{im\\theta}`

    :param theta: Array :math:`\\theta` in the formula
    :param mi: Integer or half-integer :math:`m` in the formula\\
    :return: Array of tf.complex. Same length as **theta**
    """
    theta_i = tf.reshape(theta, (-1, 1))
    mi = tf.cast(mi, dtype=theta.dtype)
    m_theta = mi * theta_i
    zeros = tf.zeros_like(m_theta)
    im_theta = tf.complex(zeros, m_theta)
    exp_theta = tf.exp(im_theta)
    return exp_theta


def D_matrix_conj(alpha, beta, gamma, j):
    """
    The conjugated D-matrix element with indices (:math:`m_1,m_2`) is

    .. math::
        D^{j}_{m_1,m_2}(\\alpha, \\beta, \\gamma)^\\star =
                        e^{i m_1 \\alpha} d^{j}_{m_1,m_2}(\\beta) e^{i m_2 \\gamma}

    :param alpha: Array
    :param beta: Array
    :param gamma: Array
    :param j: Integer :math:`2j` in the formula
    :return: Array of the conjugated D-matrices. Same shape as **alpha**, **beta**, and **gamma**
    """
    m = tf.reshape(np.arange(-j / 2, j / 2 + 1, 1), (1, -1))

    d = small_d_matrix(beta, j)
    expi_alpha = tf.reshape(exp_i(alpha, m), (-1, j + 1, 1))
    expi_gamma = tf.reshape(exp_i(gamma, m), (-1, 1, j + 1))
    expi_gamma = tf.cast(expi_gamma, dtype=expi_alpha.dtype)
    dc = tf.complex(d, tf.zeros_like(d))
    ret = tf.cast(expi_alpha * expi_gamma, dc.dtype) * dc
    return ret


def get_D_matrix_for_angle(angle, j, cached=True):
    """
    Interface to *D_matrix_conj()*

    :param angle: Dict of angle data {"alpha","beta","gamma"}
    :param j: Integer :math:`2j` in the formula
    :param cached: Haven't been used???
    :return: Array of the conjugated D-matrices. Same length as the angle data
    """
    alpha = angle["alpha"]
    beta = angle["beta"]
    gamma = angle["gamma"]
    name = "D_matrix_{}".format(j)
    if cached:
        if name not in angle:
            angle[name] = D_matrix_conj(alpha, beta, gamma, j)
        return angle[name]
    return D_matrix_conj(alpha, beta, gamma, j)


def get_D_matrix_lambda(angle, ja, la, lb, lc=None):
    """
    Get the D-matrix element

    :param angle: Dict of angle data {"alpha","beta","gamma"}
    :param ja:
    :param la:
    :param lb:
    :param lc:
    :return:
    """
    d = get_D_matrix_for_angle(angle, _spin_int(2 * ja))
    if lc is None:
        return tf.reshape(
            Dfun_delta_v2(d, ja, la, lb, (0,)), (-1, len(la), len(lb))
        )
    else:
        return Dfun_delta_v2(d, ja, la, lb, lc)



import functools
import os
import warnings

import numpy as np

from tf_pwa.amp import get_particle
from tf_pwa.cal_angle import (
    cal_angle_from_momentum,
    load_dat_file,
    parity_trans,
)
from tf_pwa.config import create_config, get_config, regist_config, temp_config
from tf_pwa.data import (
    data_index,
    data_shape,
    data_split,
    data_to_tensor,
    load_data,
    save_data,
)

DATA_MODE = "data_mode"
regist_config(DATA_MODE, {})


def register_data_mode(name=None, f=None):
    """register a data mode

    :params name: mode name used in configuration
    :params f: Data Mode class
    """

    def regist(g):
        if name is None:
            my_name = g.__name__
        else:
            my_name = name
        config = get_config(DATA_MODE)
        if my_name in config:
            warnings.warn("Override mode {}".format(my_name))
        config[my_name] = g
        return g

    if f is None:
        return regist
    return regist(f)


def load_data_mode(dic, decay_struct, default_mode="multi"):
    if dic is None:
        dic = {}
    mode = dic.get("mode", default_mode)
    return get_config(DATA_MODE)[mode](dic, decay_struct)


@register_data_mode("simple")
class SimpleData:
    def __init__(self, dic, decay_struct):
        self.decay_struct = decay_struct
        self.dic = dic
        self.cached_data = None
        chain_map = self.decay_struct.get_chains_map()
        self.re_map = {}
        for i in chain_map:
            for _, j in i.items():
                for k, v in j.items():
                    self.re_map[v] = k
        self.scale_list = self.dic.get("scale_list", ["bg"])

    def get_data_file(self, idx):
        if idx in self.dic:
            ret = self.dic[idx]
        else:
            ret = None
        return ret

    def get_dat_order(self, standard=False):
        order = self.dic.get("dat_order", None)
        if order is None:
            order = list(self.decay_struct.outs)
        else:
            order = [get_particle(str(i)) for i in order]
        if not standard:
            return order

        re_map = self.decay_struct.get_chains_map()

        def particle_item():
            for j in re_map:
                for k, v in j.items():
                    for s, l in v.items():
                        yield s, l

        new_order = []
        for i in order:
            for s, l in particle_item():
                if str(l) == str(i):
                    new_order.append(s)
                    break
            else:
                new_order.append(i)
        return new_order

    def get_weight_sign(self, idx):
        negtive_idx = self.dic.get("negtive_idx", ["bg"])
        weight_sign = 1
        if idx in negtive_idx:
            weight_sign = -1
        return weight_sign

    def get_data(self, idx) -> dict:
        if self.cached_data is not None:
            data = self.cached_data.get(idx, None)
            if data is not None:
                return data
        files = self.get_data_file(idx)
        weights = self.dic.get(idx + "_weight", None)
        weight_sign = self.get_weight_sign(idx)
        charge = self.dic.get(idx + "_charge", None)
        ret = self.load_data(files, weights, weight_sign, charge)
        return self.process_scale(idx, ret)

    def process_scale(self, idx, data):
        if idx in self.scale_list and self.dic.get("weight_scale", False):
            n_bg = data_shape(data)
            scale_factor = self.get_n_data() / n_bg
            data["weight"] = (
                data.get("weight", np.ones((n_bg,))) * scale_factor
            )
        return data

    def get_n_data(self):
        data = self.get_data("data")
        weight = data.get("weight", np.ones((data_shape(data),)))
        return np.sum(weight)

    def load_p4(self, fnames):
        particles = self.get_dat_order()
        p = load_dat_file(fnames, particles)
        return p

    def cal_angle(self, p4):
        if isinstance(p4, (list, tuple)):
            p4 = {k: v for k, v in zip(self.get_dat_order(), p4)}
        center_mass = self.dic.get("center_mass", False)
        r_boost = self.dic.get("r_boost", True)
        random_z = self.dic.get("random_z", True)
        data = cal_angle_from_momentum(
            p4,
            self.decay_struct,
            center_mass=center_mass,
            r_boost=r_boost,
            random_z=random_z,
        )
        return data

    def load_data(
        self, files, weights=None, weights_sign=1, charge=None
    ) -> dict:
        # print(files, weights)
        if files is None:
            return None
        order = self.get_dat_order()
        charges = None if charge is None else self.load_weight_file(charge)
        p4 = self.load_p4(files)
        cp_trans = self.dic.get("cp_trans", True)
        if cp_trans and charges is not None:
            p4 = {k: parity_trans(v, charges) for k, v in p4.items()}
        data = self.cal_angle(p4)
        if weights is not None:
            if isinstance(weights, float):
                data["weight"] = np.array(
                    [weights * weights_sign] * data_shape(data)
                )
            elif isinstance(weights, str):  # weight files
                weight = self.load_weight_file(weights)
                data["weight"] = weight[: data_shape(data)] * weights_sign
            else:
                raise TypeError(
                    "weight format error: {}".format(type(weights))
                )
        if charge is not None:
            data["charge_conjugation"] = charges[: data_shape(data)]
        else:
            data["charge_conjugation"] = np.ones((data_shape(data),))
        return data

    def load_weight_file(self, weight_files):
        ret = []
        if isinstance(weight_files, list):
            for i in weight_files:
                data = np.loadtxt(i).reshape((-1,))
                ret.append(data)
        elif isinstance(weight_files, str):
            data = np.loadtxt(weight_files).reshape((-1,))
            ret.append(data)
        else:
            raise TypeError(
                "weight files must be string of list of strings, not {}".format(
                    type(weight_files)
                )
            )
        if len(ret) == 1:
            return ret[0]
        return np.concatenate(ret)

    def load_cached_data(self, file_name=None):
        if file_name is None:
            file_name = self.dic.get("cached_data", None)
        if file_name is not None and os.path.exists(file_name):
            if self.cached_data is None:
                self.cached_data = load_data(file_name)
                print("load cached_data {}".format(file_name))

    def save_cached_data(self, data, file_name=None):
        if file_name is None:
            file_name = self.dic.get("cached_data", None)
        if file_name is not None:
            if not os.path.exists(file_name):
                save_data(file_name, data)
                print("save cached_data {}".format(file_name))

    def get_all_data(self):
        datafile = ["data", "phsp", "bg", "inmc"]
        self.load_cached_data()
        data, phsp, bg, inmc = [self.get_data(i) for i in datafile]
        self.save_cached_data(dict(zip(datafile, [data, phsp, bg, inmc])))
        return data, phsp, bg, inmc

    def get_data_index(self, sub, name):
        dec = self.decay_struct.topology_structure()
        if sub == "mass":
            p = get_particle(name)
            return "particle", self.re_map.get(p, p), "m"
        if sub == "p":
            p = get_particle(name)
            return "particle", self.re_map.get(p, p), "p"
        if sub == "angle":
            name_i = name.split("/")
            de_i = self.decay_struct.get_decay_chain(name_i)
            p = get_particle(name_i[-1])
            for i in de_i:
                if p in i.outs:
                    de = i
                    break
            else:
                raise IndexError("not found such decay {}".format(name))
            return (
                "decay",
                de_i.standard_topology(),
                self.re_map.get(de, de),
                self.re_map.get(p, p),
                "ang",
            )
        if sub == "aligned_angle":
            name_i = name.split("/")
            de_i = self.decay_struct.get_decay_chain(name_i)
            p = get_particle(name_i[-1])
            for i in de_i:
                if p in i.outs:
                    de = i
                    break
            else:
                raise IndexError("not found such decay {}".format(name))
            return (
                "decay",
                de_i.standard_topology(),
                self.re_map.get(de, de),
                self.re_map.get(p, p),
                "aligned_angle",
            )
        raise ValueError("unknown sub {}".format(sub))

    def get_phsp_noeff(self):
        if "phsp_noeff" in self.dic:
            phsp_noeff = self.get_data("phsp_noeff")
            return phsp_noeff
        warnings.warn(
            "No data file as 'phsp_noeff', using the first 'phsp' file instead."
        )
        return self.get_data("phsp")

    def get_phsp_plot(self):
        if "phsp_plot" in self.dic:
            return self.get_data("phsp_plot")
        return self.get_data("phsp")


@register_data_mode("multi")
class MultiData(SimpleData):
    def __init__(self, *args, **kwargs):
        super(MultiData, self).__init__(*args, **kwargs)
        self._Ngroup = 0

    def process_scale(self, idx, data):
        if idx in self.scale_list and self.dic.get("weight_scale", False):
            for i, data_i in enumerate(data):
                n_bg = data_shape(data_i)
                scale_factor = self.get_n_data()[i] / n_bg
                data_i["weight"] = (
                    data_i.get("weight", np.ones((n_bg,))) * scale_factor
                )
        return data

    def get_n_data(self):
        data = self.get_data("data")
        weight = [
            data_i.get("weight", np.ones((data_shape(data_i),)))
            for data_i in data
        ]
        return [np.sum(weight_i) for weight_i in weight]

    @functools.lru_cache()
    def get_data(self, idx) -> list:
        if self.cached_data is not None:
            data = self.cached_data.get(idx, None)
            if data is not None:
                return data
        files = self.get_data_file(idx)
        if files is None:
            return None
        if not isinstance(files[0], list):
            files = [files]
        weights = self.dic.get(idx + "_weight", None)
        if weights is None:
            weights = [None] * len(files)
        elif isinstance(weights, float):
            weights = [weights] * len(files)
        weight_sign = self.get_weight_sign(idx)
        charge = self.dic.get(idx + "_charge", None)
        if charge is None:
            charge = [None] * len(files)
        elif not isinstance(charge[0], list):
            charge = [charge]
        ret = [
            self.load_data(i, j, weight_sign, k)
            for i, j, k in zip(files, weights, charge)
        ]
        if self._Ngroup == 0:
            self._Ngroup = len(ret)
        elif idx != "phsp_noeff":
            assert self._Ngroup == len(ret), "not the same data group"
        bg_value = self.dic.get(idx + "_bg_value", None)
        if bg_value is not None:
            if isinstance(bg_value, str):
                bg_value = [bg_value]
            for i, file_name in enumerate(bg_value):
                ret[i]["bg_value"] = np.reshape(np.loadtxt(file_name), (-1,))
        eff_value = self.dic.get(idx + "_eff_value", None)
        if eff_value is not None:
            if isinstance(eff_value, str):
                eff_value = [eff_value]
            for i, file_name in enumerate(eff_value):
                ret[i]["eff_value"] = np.reshape(np.loadtxt(file_name), (-1,))
        ret = self.process_scale(idx, ret)
        return ret

    def get_phsp_noeff(self):
        if "phsp_noeff" in self.dic:
            phsp_noeff = self.get_data("phsp_noeff")
            assert len(phsp_noeff) == 1
            return phsp_noeff[0]
        warnings.warn(
            "No data file as 'phsp_noeff', using the first 'phsp' file instead."
        )
        return self.get_data("phsp")[0]



from contextlib import contextmanager
from functools import partial


class ConfigManager(dict):
    pass


class _Default:
    pass


default_var = _Default()


def create_config(default=None):
    if default is None:
        default = {}
    _config = ConfigManager(default)

    def set_(name, var):
        """
        set a configuration.
        """
        if name in _config:
            _config[name] = var
        else:
            raise Exception("No configuration named {} found.".format(name))

    def get_(name, default=default_var):
        """
        get a configuration.
        """
        if name in _config:
            return _config[name]
        if isinstance(default, _Default):
            raise Exception("No configuration named {} found.".format(name))
        return default

    def regist_(name, var=None):
        """
        regist a configuration.
        """
        if name in _config:
            raise Exception(
                "Configuration named {} already exists.".format(name)
            )
        if var is None:

            def regist(f):
                _config[name] = f
                return f

            return regist
        _config[name] = var
        return var

    return set_, get_, regist_


set_config, get_config, regist_config = create_config(
    {"multi_gpus": False, "dtype": "float64", "complex_dtype": "complex128"}
)


@contextmanager
def temp_config(name, var):
    tmp = get_config(name)
    set_config(name, var)
    yield var
    set_config(name, tmp)


using_amplitude = lambda var: temp_config("amp", var)



from scipy.stats import chi2, norm


def prob(chi_2, ndf):
    """
    Computation of the probability for a certain Chi-squared (chi2)
    and number of degrees of freedom (ndf).

    Calculations are based on the incomplete gamma function P(a,x),
    where a=ndf/2 and x=chi2/2.

    P(a,x) represents the probability that the observed Chi-squared
    for a correct model should be less than the value chi2.

    The returned probability corresponds to 1-P(a,x),
    which denotes the probability that an observed Chi-squared exceeds
    the value chi2 by chance, even for a correct model.

    """
    if ndf <= 0.0:
        return 0.0  # Set CL(confidential level) to zero in case ndf<=0
    if chi_2 <= 0.0:
        if chi_2 < 0.0:
            return 0.0
        else:
            return 1.0
    return chi2.sf(chi_2, ndf)


def erfc_inverse(x):
    """ erfc-1(x) = - 1/sqrt(2) * normal_quantile( 0.5 * x)"""
    return -0.70710678118654752440 * normal_quantile(0.5 * x)


def normal_quantile(p):
    """
    Computes quantiles for standard normal distribution N(0, 1)
    at probability p

    """
    if p < 0.0 or p > 1.0:
        raise Exception("probability outside (0, 1)")
        return 0.0
    return norm.ppf(p)


def significance(l1: float, l2: float, ndf: int) -> float:
    """
    computation of significance for log-likelihood fit values l1 and l2
    with number of degrees of freedom (ndf).
    """
    DeltaLL = 2.0 * abs(l1 - l2)
    p = prob(DeltaLL, ndf)
    # math.sqrt(2) * erfc_inverse(p)
    return -normal_quantile(p / 2.0)



import os.path
import sys

import matplotlib.pyplot as plt
import numpy as np
import uproot

from tf_pwa.angle import LorentzVector as lv
from tf_pwa.angle import Vector3 as v3
from tf_pwa.phasespace import PhaseSpaceGenerator

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def generate_mc(num):
    M_pipm = 0.1395702
    M_pi0 = 0.1349766
    M_Kpm = 0.49368
    M_Dpm = 1.86961
    M_D0 = 1.86483
    M_Dstarpm = 2.01026
    M_Bpm = 5.27926
    M_B0 = 5.27964
    a = PhaseSpaceGenerator(M_Bpm, [M_Dstarpm, M_Dpm, M_Kpm])
    b = PhaseSpaceGenerator(M_Dstarpm, [M_D0, M_pipm])
    p_D1, p_D2, p_K = a.generate(num)
    p_D, p_pi = b.generate(num)
    return [i.numpy() for i in [p_D1, p_D2, p_K, p_D, p_pi]]


def index_bin(x, xi):
    xi1 = np.expand_dims(xi, axis=-1)
    mask = (x[:, 0] < xi1) & (x[:, 1] > xi1)
    idx1, idx2 = np.nonzero(mask)
    idx = np.zeros_like(xi, dtype="int64")
    idx[idx1] = idx2
    return idx


class EffWeight:
    def __init__(self, root_file):
        self.f = uproot.open(root_file)
        self.eff_bin = [
            self.f.get("RegDalitzEfficiency_bin{}".format(i)) for i in range(5)
        ]
        self.x_bins, self.y_bins = self.eff_bin[0].bins  # assert all bins same
        self.values = np.array([i.values for i in self.eff_bin])

    def eff_weight(self, cos_hel, m2_13, m2_23):
        n_histo = (3 - np.floor((cos_hel + 1) / 0.5) + 1).astype("int64")
        x_idx = index_bin(self.x_bins, m2_13)
        y_idx = index_bin(self.y_bins, m2_23)
        return self.values[n_histo, x_idx, y_idx]


def generate_mc_eff(num, eff_file):
    p_D1, p_D2, p_K, p_D, p_pi = generate_mc(num)
    m2_13 = lv.M2(p_D1 + p_K)
    m2_23 = lv.M2(p_D2 + p_K)
    p_D1_r = lv.rest_vector(p_D1 + p_D2, p_D1)
    cos_theta = v3.cos_theta(lv.vect(p_D1_r), lv.vect(p_D))
    eff = EffWeight(eff_file)
    weight = eff.eff_weight(cos_theta, m2_13, m2_23)
    rnd = np.random.random(num)
    mask = weight > rnd
    # plt.hist(cos_theta.numpy()[mask], bins=100)
    # plt.savefig("cos_theta.png")
    p_D1, p_D2, p_K, p_D, p_pi = [
        i[mask] for i in [p_D1, p_D2, p_K, p_D, p_pi]
    ]
    return p_D1, p_D2, p_K, p_D, p_pi


def main(num, eff_file):
    p_D1, p_D2, p_K, p_D, p_pi = generate_mc_eff(num, eff_file)
    p_D, p_pi = [lv.rest_vector(lv.neg(p_D1), i) for i in [p_D, p_pi]]
    data = np.array([p_D1, p_D2, p_K])
    data = np.transpose(data, (1, 0, 2))
    data2 = np.array([p_D, p_pi])
    data2 = np.transpose(data2, (1, 0, 2))
    np.savetxt("data/gen_mc.dat", data.reshape(-1, 4))
    np.savetxt("data/gen_mc_Dstar.dat", data2.reshape(-1, 4))
    bins, x, y, _ = plt.hist2d(lv.M2(p_D1 + p_K), lv.M2(p_D2 + p_K), bins=50)
    plt.clf()
    plt.contourf(
        *np.meshgrid(x[1:] / 2 + x[:-1] / 2, y[1:] / 2 + y[:-1] / 2),
        bins.astype("float"),
    )
    plt.colorbar()
    print(lv.M(p_D1 + p_D2 + p_K))
    plt.savefig("m13_m23.png")


if __name__ == "__main__":
    main(1000000, "data/eff/Efficiency_BmDstpDmKm_run2.root")



import numpy as np

from tf_pwa.histogram import *


def test_hist1d():
    data = np.linspace(0, 1, 500)
    weight = np.cos((data - 0.5) * np.pi)
    hist = Hist1D.histogram(data, weights=weight, bins=30)
    plt.clf()
    ax = plt.subplot2grid((4, 1), (0, 0), rowspan=3)
    hist.draw(ax)
    hist.draw_line()
    hist.draw_line(kind="quadratic")
    hist.draw_kde(ax, kind="gauss", color="blue")
    hist.draw_kde(ax, kind="cauchy", color="red")
    (0.1 * hist + hist * 0.1).draw_bar(ax)
    hist.draw_error(ax)
    hist2 = Hist1D.histogram(
        data,
        weights=weight
        + np.random.random(
            500,
        )
        * 0.1
        - 0.05,
        bins=30,
    )
    ax2 = plt.subplot2grid((4, 1), (3, 0), rowspan=1)
    (hist2 - hist).draw_pull(ax2)
    plt.savefig("hist1d_test1.png")


def test_weighteddata():
    data = np.linspace(0, 1, 500)
    weight = np.cos((data - 0.5) * np.pi)
    hist = WeightedData(data, weights=weight, bins=30)
    plt.clf()
    hist.draw()
    hist.draw_kde(kind="gauss", color="blue")
    hist.draw_kde(kind="uniform", color="green")
    hist.draw_kde(kind="epanechnikov", color="red")
    hist.draw_kde(kind=gauss, color="pink")
    hist2 = hist * 0.1 + hist
    plt.savefig("hist1d_test2.png")


def test_weight():
    data = np.linspace(0, 1, 500)
    weight = np.cos((data - 0.5) * np.pi)
    hist1 = Hist1D.histogram(data, weights=weight, bins=30)
    hist2 = Hist1D.histogram(data, bins=30)
    a, b = np.histogram(data, weights=weight)
    hist3 = Hist1D(b, a)
    hist4 = WeightedData(data, bins=30)
    assert hist1.get_count() == hist3.get_count()
    assert hist2.get_count() == hist4.get_count()
    assert hist1.scale_to(hist2) == 1 / np.mean(weight)



import numpy as np

from tf_pwa.adaptive_bins import AdaptiveBound
from tf_pwa.adaptive_bins import cal_chi2 as cal_chi2_o
from tf_pwa.data import (
    data_index,
    data_merge,
    data_shape,
    data_split,
    data_to_numpy,
    load_data,
    save_data,
)

from .config_loader import ConfigLoader


@ConfigLoader.register_function()
def cal_bins_numbers(
    self, adapter, data, phsp, read_data, bg=None, bg_weight=None
):
    data_cut = read_data(data)
    # print(data_cut)
    amp_weight = self.get_amplitude()(phsp).numpy()
    phsp_cut = read_data(
        phsp
    )  # np.array([data_index(phsp, idx) for idx in data_idx])
    phsp_slice = np.concatenate([phsp_cut, [amp_weight]], axis=0)
    phsps = adapter.split_data(phsp_slice)
    datas = adapter.split_data(data_cut)
    bound = adapter.get_bounds()
    weight_scale = self.config["data"].get("weight_scale", False)
    if bg is not None:
        bg_cut = read_data(
            bg
        )  # np.array([data_index(bg, idx) for idx in data_idx])
        bgs = adapter.split_data(bg_cut)
        if weight_scale:
            int_norm = (
                data_cut.shape[-1] * (1 - bg_weight) / np.sum(amp_weight)
            )
        else:
            int_norm = (
                data_cut.shape[-1] - bg_cut.shape[-1] * bg_weight
            ) / np.sum(amp_weight)
    else:
        int_norm = data_cut.shape[-1] / np.sum(amp_weight)
    # print("int norm:", int_norm)
    ret = []
    for i, bnd in enumerate(bound):
        min_x, min_y = bnd[0]
        max_x, max_y = bnd[1]
        ndata = datas[i].shape[-1]
        print(ndata)
        nmc = np.sum(phsps[i][2]) * int_norm
        if bg is not None:
            nmc += bgs[i].shape[-1] * bg_weight
        n_exp = nmc
        ret.append((ndata, nmc))
    return ret


@ConfigLoader.register_function()
def cal_chi2(self, read_data=None, bins=[[2, 2]] * 3, mass=["R_BD", "R_CD"]):
    if read_data is None:
        data_idx = [self.get_data_index("mass", i) for i in mass]
        read_data = lambda a: np.array(
            [data_index(a, idx) ** 2 for idx in data_idx]
        )
    data, bg, phsp, _ = self.get_all_data()
    bg_weight = self._get_bg_weight()[0]
    all_data = data_merge(*data)
    all_data_cut = read_data(
        all_data
    )  # np.array([data_index(a, idx) for idx in data_idx])
    adapter = AdaptiveBound(all_data_cut, bins)
    numbers = [
        self.cal_bins_numbers(adapter, i, j, read_data, k, l)
        for i, j, k, l in zip(data, phsp, bg, bg_weight)
    ]
    numbers = np.array(numbers)
    numbers = np.sum(numbers, axis=0)
    return cal_chi2_o(numbers, self.get_ndf())



import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import UnivariateSpline, interp1d


def plot_hist(binning, count, ax=plt, **kwargs):
    n = count.shape[0]
    a = np.zeros((n + 2,))
    b = np.zeros((n + 2,))
    a[:-1] = binning
    a[-1] = binning[-1] + (binning[-1] - binning[-2]) / 20
    b[1:-1] = count
    return ax.step(a, b, **kwargs)


def interp_hist(binning, y, num=1000, kind="UnivariateSpline"):
    """interpolate data from hostgram into a line"""
    x = (binning[:-1] + binning[1:]) / 2
    if kind == "UnivariateSpline":
        func = UnivariateSpline(x, y, s=2)
    else:
        func = interp1d(x, y, kind=kind, fill_value="extrapolate")
    x_new = np.linspace(
        np.min(binning), np.max(binning), num=num, endpoint=True
    )
    y_new = func(x_new)
    return x_new, y_new


def gauss(x):
    return np.exp(-(x ** 2) / 2) / np.sqrt(2 * np.pi)


def cauchy(x):
    return 1 / (x ** 2 + 1) / np.pi


def epanechnikov(x):
    return np.where((x < 1) & (x > -1), (1 - x ** 2) / 4 * 3, 0)


def uniform(x):
    return np.where((x < 1) & (x > -1), 0.5, 0)


def weighted_kde(m, w, bw, kind="gauss"):
    n = w.shape[0]

    kind_map = {
        "gauss": gauss,
        "cauchy": cauchy,
        "epanechnikov": epanechnikov,
        "uniform": uniform,
    }
    if isinstance(kind, str):
        kernel = kind_map[kind]
    else:
        kernel = kind

    def f(x):
        ret = np.zeros_like(x)
        for i in range(n):
            y = (x - m[i]) / bw[i]
            tmp = w[i] * kernel(y)
            ret += tmp
        return ret

    return f


class Hist1D:
    def __init__(self, binning, count, error=None):
        if error is None:
            error = np.sqrt(count)
        self.binning = binning
        self.count = count
        self.error = error
        self._cached_color = None

    def draw(self, ax=plt, **kwargs):
        a = plot_hist(self.binning, self.count, ax=ax, **kwargs)
        self._cached_color = a[0].get_color()
        return a

    def draw_bar(self, ax=plt, **kwargs):
        return ax.bar(
            self.bin_center,
            self.count,
            width=self.bin_width,
            **kwargs,
        )

    def draw_kde(self, ax=plt, kind="gauss", bin_scale=1.0, **kwargs):
        color = kwargs.pop("color", self._cached_color)
        m = self.bin_center
        bw = self.bin_width * bin_scale
        kde = weighted_kde(m, self.count, bw, kind)
        x = np.linspace(
            self.binning[0], self.binning[-1], self.count.shape[0] * 10
        )
        if "fmt" in kwargs:
            fmt = kwargs.pop("fmt")
            return ax.plot(x, kde(x), fmt, color=color, **kwargs)
        else:
            return ax.plot(x, kde(x), color=color, **kwargs)

    def draw_pull(self, ax=plt, **kwargs):
        with np.errstate(divide="ignore", invalid="ignore"):
            y_error = np.where(self.error == 0, 0, self.count / self.error)
        return ax.bar(
            self.bin_center,
            y_error,
            width=self.bin_width,
            **kwargs,
        )

    def draw_line(self, ax=plt, num=1000, kind="UnivariateSpline", **kwargs):
        x_new, y_new = interp_hist(self.binning, self.count, num, kind)
        return ax.plot(x_new, y_new, **kwargs)

    def draw_error(self, ax=plt, fmt="none", **kwargs):
        color = kwargs.pop("color", self._cached_color)
        return ax.errorbar(
            self.bin_center,
            y=self.count,
            xerr=self.bin_width / 2,
            yerr=self.error,
            fmt=fmt,
            color=color,
            **kwargs,
        )

    @property
    def bin_center(self):
        return (self.binning[:-1] + self.binning[1:]) / 2

    @property
    def bin_width(self):
        return self.binning[1:] - self.binning[:-1]

    def get_bin_weight(self):
        return (self.binning[-1] - self.binning[0]) / (
            self.binning.shape[0] - 1
        )

    def __mul__(self, other):
        if isinstance(other, (float, int)):
            return Hist1D(self.binning, self.count * other, self.error * other)
        raise NotImplementedError

    __rmul__ = __mul__

    def __add__(self, other):
        assert np.allclose(
            self.binning, other.binning
        ), "need to be the same binning"
        return Hist1D(
            self.binning,
            self.count + other.count,
            np.sqrt(self.error ** 2 + other.error ** 2),
        )

    def __sub__(self, other):
        assert np.allclose(
            self.binning, other.binning
        ), "need to be the same binning"
        return Hist1D(
            self.binning,
            self.count - other.count,
            np.sqrt(self.error ** 2 + other.error ** 2),
        )

    @staticmethod
    def histogram(m, *args, weights=None, **kwargs):
        if weights is None:
            count, binning = np.histogram(m, *args, **kwargs)
            count2, _ = np.histogram(m, *args, **kwargs)
        else:
            weights = np.asarray(weights)
            count, binning = np.histogram(m, *args, weights=weights, **kwargs)
            count2, _ = np.histogram(m, *args, weights=weights ** 2, **kwargs)
        return Hist1D(binning, count, np.sqrt(count2))

    def scale_to(self, other):
        scale_factor = other.get_count() / self.get_count()
        bin_width_factor = np.mean(other.bin_width) / np.mean(self.bin_width)
        scale = scale_factor * bin_width_factor
        self.count *= scale
        self.error *= scale
        return scale

    def get_count(self):
        return np.sum(self.count)


class WeightedData(Hist1D):
    def __init__(self, m, *args, weights=None, **kwargs):
        if weights is None:
            weights = np.ones_like(m)
        count, binning = np.histogram(m, *args, weights=weights, **kwargs)
        count2, _ = np.histogram(m, *args, weights=weights ** 2, **kwargs)
        self.value = m
        self.weights = weights
        super().__init__(binning, count, np.sqrt(count2))

    def draw_kde(self, ax=plt, kind="gauss", bin_scale=1.0, **kwargs):
        color = kwargs.pop("color", self._cached_color)
        bw = np.mean(self.bin_width) * bin_scale * np.ones_like(self.value)
        kde = weighted_kde(self.value, self.weights, bw, kind)
        x = np.linspace(
            self.binning[0], self.binning[-1], self.count.shape[0] * 10
        )
        return ax.plot(x, kde(x), color=color, **kwargs)

    def __add__(self, other):
        assert np.allclose(
            self.binning, other.binning
        ), "need to be the same binning"
        ret = WeightedData(
            np.concatenate([self.value, other.value]),
            weights=np.concatenate([self.weights, other.weights]),
        )
        ret.binning = self.binning
        ret.count = self.count + other.count
        ret.error = np.sqrt(self.error ** 2 + other.error ** 2)
        return ret

    def __mul__(self, other):
        if isinstance(other, (float, int)):
            ret = WeightedData(self.value, weights=self.weights * other)
            ret.binning = self.binning
            ret.error = self.error * other
            ret.count = self.count * other
            return ret
        raise NotImplementedError

    def scale_to(self, other):
        scale = super().scale_to(other)
        self.weights *= scale
        return scale



import yaml

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.tests.common import write_temp_file

config_str = """

data:
    data: "data.dat"
    phsp: "PHSP.dat"

decay:
    ee:
    - [R1, b]
    - - R2
      - c
    - [R3, d, model: default]
    R1: [c, d, l_list: [1]]
    R2:
      - b
      - d
    R3: [b, c]

particle:
    $top:
        ee:
            J: 1
            P: -1
            spins: [-1, 1]
    $finals:
        b: {{J: 1, P: -1}}
        c: {{J: 1, P: -1}}
        d: {{J: 0, P: -1}}
    R1: [R1_a, R1_b]
    R2: {{J: 1, P: -1, mass: 2.3, width: 0.03}}
    R3: {{J: 1, P: -1, mass: 2.3, width: 0.03}}
    $include: "{file_name}"

"""

resonancs_str = """

R1_a: {J: 1, P: -1, mass: 2.3, width: 0.03}
R1_b: {J: 1, P: -1, mass: 2.3, width: 0.03}

"""


def test_load():
    with write_temp_file(resonancs_str) as f:
        cs = config_str.format(file_name=f)
        print(cs)
        with write_temp_file(cs) as g:
            config = ConfigLoader(g)
            with open(g) as f:
                data = yaml.full_load(f)
            config2 = ConfigLoader(data)
    config.get_amplitude()
    config2.get_amplitude()


def test_constrains():
    with write_temp_file(resonancs_str) as f:
        cs = config_str.format(file_name=f)
        print(cs)
        with write_temp_file(cs) as g:
            config = ConfigLoader(g)

    amp = config.get_amplitude()
    config.add_free_var_constraints(amp)



import contextlib
import os
import tempfile


@contextlib.contextmanager
def write_temp_file(s, filename=None):
    if filename is None:
        a = tempfile.mktemp()
    else:
        a = filename
    with open(a, "w") as f:
        f.write(s)
    yield a
    os.remove(a)



import os.path
import re
import sys

import yaml

from tf_pwa.dec_parser import load_dec_file

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


commands = {}
config = {"particle": {}, "decay": {}}
model_map = {"HELCOV": "default"}


def register_commands(name):
    def regist_(f):
        commands[name] = f

    return regist_


@register_commands("Particle")
def add_particle(params):
    particle = config["particle"]
    ret = {}
    params_name = ["mass", "width"]
    for k, v in zip(params_name, params["params"]):
        ret[k] = float(v)
    particle[params["name"]] = ret


@register_commands("RUNNINGWIDTH")
def add_running_width(params):
    for i in params:
        config["particle"][i]["running_width"] = True


@register_commands("Decay")
def add_decay(params):
    name = params["name"]
    finals = params["final"]
    ret = []
    for i in finals:
        kwargs = i.copy()
        outs = kwargs.pop("outs")
        for j in ["total"]:
            kwargs.pop(j)
        model = kwargs.get("model", "default")
        if model in model_map:
            kwargs["model"] = model_map[model]
        ret.append(outs + [kwargs])
    if ret:
        config["decay"][name] = ret


def load_parity_list(parity_list="_parity.list_"):
    patten = re.compile(r"([^\s]*)\s+([+-])(\d+)")
    with open(parity_list) as f:
        for i in f.readlines():
            grp = patten.match(i)
            if grp:
                name = grp[1]
                p = int(grp[2] + "1")
                j = int(grp[3])
                if name in config["particle"]:
                    config["particle"][name]["J"] = j
                    config["particle"][name]["P"] = p


def dec2yml(in_file, out_file, parity_list="_parity.list_"):
    with open("9R_c.dec") as f:
        dec = load_dec_file(f)
    for i in dec:
        # print(i)
        commands[i[0]](i[1])
    try:
        load_parity_list(parity_list)
    except FileNotFoundError:
        pass
    with open("config_test.yml", "w") as f:
        yaml.dump(config, f)


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("in_file")
    parser.add_argument("-o", dest="out_file", default="config.yml")
    parser.add_argument(
        "--parity", dest="parity_file", default="_parity.list_"
    )
    results = parser.parse_args()
    dec2yml(results.in_file, results.out_file, results.parity_file)


if __name__ == "__main__":
    main()



import argparse
import inspect
import sys
import warnings

__all__ = ["regist_subcommand"]


def _protect_dict(override_message=None):
    d = {}

    def get_dict():
        return d.copy()

    def set_var(name, var):
        if name in d:
            if override_message is not None:
                warnings.warn("override {}: {}".format(override_message, name))
        d[name] = var

    return get_dict, set_var


get_sub_cmd, set_sub_cmd = _protect_dict("sub commands")


def regist_subcommand(name=None, arg_fun=None, args=None):
    _sub_commands = get_sub_cmd()
    if args is None:
        args = {}

    def wrap(f):
        name_t = name
        if name_t is None:
            name_t = f.__name__
        if arg_fun is None:
            cmds = _build_arguments(f, args)
        else:
            cmds = arg_fun
        set_sub_cmd(name_t, cmds)
        return f

    return wrap


def _build_arguments(f, config_args):
    argspec = inspect.getfullargspec(f)

    def wrap_f(arg):
        args = []
        kwargs = {}
        for i in argspec.args:
            if hasattr(arg, i):
                var = getattr(arg, i)
                args.append(var)
        if argspec.kwonlyargs:
            for i in argspec.kwonlyargs:
                if hasattr(arg, i):
                    var = getattr(arg, i)
                    kwargs[i] = var
        return f(*args, **kwargs)

    wrap_f.__name__ = f.__name__

    ret = {"fun": wrap_f, "args": []}

    arg_defaults = {}
    if argspec.defaults:
        for i, j in zip(argspec.defaults[::-1], argspec.args[::-1]):
            arg_defaults[j] = i
    if argspec.kwonlydefaults:
        for i in argspec.kwonlydefaults:
            arg_defaults[i] = argspec.kwonlydefaults[i]

    arg_type = {}
    if hasattr(argspec, "annotations"):
        arg_type = argspec.annotations

    for i in config_args:
        ret["args"].append(config_args[i])

    for i in argspec.args:
        if i in config_args:
            break
        args = (i,)
        kwargs = {}
        if i in arg_defaults:
            kwargs["nargs"] = "?"
            kwargs["default"] = arg_defaults[i]
        if i in arg_type:
            kwargs["type"] = arg_type[i]
        ret["args"].append((args, kwargs))

    if not argspec.kwonlyargs:
        return ret

    for i in argspec.kwonlyargs:
        if i in config_args:
            break
        name = "--" + i
        args = (name,)
        kwargs = {}
        if i in arg_defaults:
            kwargs["default"] = arg_defaults[i]
        if i in arg_type:
            kwargs["type"] = arg_type[i]
        ret["args"].append((args, kwargs))

    return ret


@regist_subcommand(name="help")
def help_function():
    print(
        """
    using ```python -m tf_pwa [subprocess]```
    """
    )


def main(argv=None):
    parser = argparse.ArgumentParser("tf_pwa")
    subparsers = parser.add_subparsers()
    _sub_commands = get_sub_cmd()
    for i in _sub_commands:
        cmds = _sub_commands[i]
        pi = subparsers.add_parser(i)
        if callable(cmds):
            cmds(pi)
        else:
            pi.set_defaults(func=cmds["fun"])
            for args, kwargs in cmds["args"]:
                pi.add_argument(*args, **kwargs)
    rargs = parser.parse_args(argv)
    if hasattr(rargs, "func"):

        return rargs.func(rargs)
    else:
        help_function()



"""
This module provides functions to describe the lineshapes of the intermediate particles, namely generalized
Breit-Wigner function. Users can also define new lineshape using the function wrapper **regist_lineshape()**.
"""

import fractions
import functools
import math
import warnings

import sympy as sym

from .tensorflow_wrapper import tf

breit_wigner_dict = {}


def regist_lineshape(name=None):
    """
    It will be used as a wrapper to define various Breit-Wigner functions

    :param name: String name of the BW function
    :return: A function used in a wrapper
    """

    def fopt(f):
        name_t = name
        if name_t is None:
            name_t = f.__name__
        if name_t in breit_wigner_dict:
            warnings.warn(
                "override breit wigner function :", name
            )  # warning to users
        breit_wigner_dict[name_t] = f  # function
        return f

    return fopt


@regist_lineshape("one")
def one(*args):
    """
    A uniform function
    """
    return tf.complex(
        1.0, 0.0
    )  # breit_wigner_dict["one"]==tf.complex(1.0,0.0)


@regist_lineshape("BW")
def BW(m, m0, g0, *args):
    """
    Breit-Wigner function

    .. math::
        BW(m) = \\frac{1}{m_0^2 - m^2 -  i m_0 \\Gamma_0 }

    """
    m0 = tf.cast(m0, m.dtype)
    gamma = tf.cast(g0, m.dtype)
    x = m0 * m0 - m * m
    y = m0 * gamma
    s = x * x + y * y
    ret = tf.complex(x / s, y / s)
    return ret


@regist_lineshape("default")  # 两个名字
@regist_lineshape("BWR")  # BW with running width
def BWR(m, m0, g0, q, q0, L, d):
    """
    Relativistic Breit-Wigner function (with running width). It's also set as the default lineshape.

    .. math::
        BW(m) = \\frac{1}{m_0^2 - m^2 -  i m_0 \\Gamma(m)}

    """
    gamma = Gamma(m, g0, q, q0, L, m0, d)
    num = 1.0
    m0 = tf.cast(m0, m.dtype)
    x = m0 * m0 - m * m
    y = m0 * gamma
    s = x * x + y * y
    ret = tf.complex(x / s, y / s)
    return ret


# added by xiexh for GS model rho
def twoBodyCMmom(m_0, m_1, m_2):
    """relative momentum for 0 -> 1 + 2"""
    M12S = m_1 + m_2
    M12D = m_1 - m_2
    if hasattr(M12S, "dtype"):
        m_0 = tf.convert_to_tensor(m_0, dtype=M12S.dtype)
    #    m_eff = tf.where(m_0 > M12S, m_0, M12S)
    #    p = (m_eff - M12S) * (m_eff + M12S) * (m_eff - M12D) * (m_eff + M12D)
    # if p is negative, which results from bad data, the return value is 0.0
    # print("p", tf.where(p==0), m_0, m_1, m_2)
    p = (m_0 - M12S) * (m_0 + M12S) * (m_0 - M12D) * (m_0 + M12D)
    zeros = tf.zeros_like(m_0)
    ret = tf.where(p > 0, tf.sqrt(p) / (2 * m_0), zeros)
    return ret


def hFun(s, daug2Mass, daug3Mass):
    _pi = 3.14159265359
    _pi = tf.cast(_pi, s.dtype)

    sm = daug2Mass + daug3Mass
    sqrt_s = tf.sqrt(s)
    k_s = twoBodyCMmom(s, daug2Mass, daug3Mass)

    ret = (
        (2.0 / _pi)
        * (k_s / sqrt_s)
        * tf.math.log((sqrt_s + 2.0 * k_s) / (sm), name="log")
    )
    ret = tf.cast(ret, s.dtype)
    return ret


def dh_dsFun(s, daug2Mass, daug3Mass):
    _pi = 3.14159265359
    _pi = tf.cast(_pi, s.dtype)
    k_s = twoBodyCMmom(s, daug2Mass, daug3Mass)

    ret = hFun(s, daug2Mass, daug3Mass) * (
        1.0 / (8.0 * tf.pow(k_s, 2)) - 1.0 / (2.0 * s)
    ) + 1.0 / (2.0 * _pi * s)
    ret = tf.cast(ret, s.dtype)
    return ret


def dFun(s, daug2Mass, daug3Mass):
    _pi = 3.14159265359
    _pi = tf.cast(_pi, s.dtype)
    sm = daug2Mass + daug3Mass
    sm24 = sm * sm / 4.0
    m = tf.sqrt(s)
    k_m2 = twoBodyCMmom(s, daug2Mass, daug3Mass)

    ret = (
        3.0
        / _pi
        * sm24
        / tf.pow(k_m2, 2)
        * tf.math.log((m + 2 * k_m2) / sm, name="log")
        + m / (2 * _pi * k_m2)
        - sm24 * m / (_pi * tf.pow(k_m2, 3))
    )
    ret = tf.cast(ret, s.dtype)
    return ret


def fsFun(s, m2, gam, daug2Mass, daug3Mass):
    k_s = twoBodyCMmom(s, daug2Mass, daug3Mass)
    k_Am2 = twoBodyCMmom(m2, daug2Mass, daug3Mass)

    f = gam * m2 / tf.pow(k_Am2, 3)
    f *= tf.pow(k_s, 2) * (
        hFun(s, daug2Mass, daug3Mass) - hFun(m2, daug2Mass, daug3Mass)
    ) + (m2 - s) * tf.pow(k_Am2, 2) * dh_dsFun(m2, daug2Mass, daug3Mass)

    f = tf.cast(f, s.dtype)
    return f


# Gounaris-Sakurai model for rho
def GS(m, m0, g0, q, q0, L, d, c_daug2Mass=0.13957039, c_daug3Mass=0.1349768):
    gamma = Gamma(m, g0, q, q0, L, m0, d)
    c_daug2Mass = tf.cast(c_daug2Mass, m.dtype)
    c_daug3Mass = tf.cast(c_daug3Mass, m.dtype)

    D = 1.0 + dFun(m0 * m0, c_daug2Mass, c_daug3Mass) * g0 / m0
    E = m0 * m0 - m * m + fsFun(m * m, m0 * m0, g0, c_daug2Mass, c_daug3Mass)
    F = m0 * gamma

    D /= E * E + F * F
    ret = tf.complex(D * E, D * F)

    return ret


# added by xiexh end


def BWR2(m, m0, g0, q2, q02, L, d):
    """
    Relativistic Breit-Wigner function (with running width). Allow complex :math:`\\Gamma`.

    .. math::
        BW(m) = \\frac{1}{m_0^2 - m^2 -  i m_0 \\Gamma(m)}

    """
    gamma = Gamma2(m, g0, q2, q02, L, m0, d)
    num = 1.0
    m0 = tf.cast(m0, m.dtype)
    x = tf.cast(m0 * m0 - m * m, gamma.dtype)
    y = tf.cast(m0, gamma.dtype) * gamma
    d = x - 1j * y
    bw_x = tf.math.real(d)
    bw_y = tf.math.imag(d)
    bw_r2 = bw_x * bw_x + bw_y * bw_y
    ret = tf.complex(bw_x / bw_r2, bw_y / bw_r2)
    return ret


def BWR_normal(m, m0, g0, q2, q02, L, d):
    """
    Relativistic Breit-Wigner function (with running width) with a normal factor.

    .. math::
        BW(m) = \\frac{\\sqrt{m_0 \\Gamma(m)}}{m_0^2 - m^2 -  i m_0 \\Gamma(m)}

    """
    gamma = Gamma2(m, g0, q2, q02, L, m0, d)
    num = 1.0
    m0 = tf.cast(m0, m.dtype)
    x = tf.cast(m0 * m0 - m * m, gamma.dtype)
    y = tf.cast(m0, gamma.dtype) * gamma
    ret = tf.sqrt(tf.cast(m0, gamma.dtype) * gamma) / (x - 1j * y)
    return ret


def Gamma(m, gamma0, q, q0, L, m0, d):
    """
    Running width in the RBW

    .. math::
        \\Gamma(m) = \\Gamma_0 \\left(\\frac{q}{q_0}\\right)^{2L+1}\\frac{m_0}{m} B_{L}'^2(q,q_0,d)

    """
    q0 = tf.cast(q0, q.dtype)
    _epsilon = 1e-15
    qq0 = tf.where(q0 > _epsilon, (q / q0) ** (2 * L + 1), 1.0)
    mm0 = tf.cast(m0, m.dtype) / m
    bp = Bprime(L, q, q0, d) ** 2
    gammaM = gamma0 * qq0 * mm0 * tf.cast(bp, qq0.dtype)
    return gammaM


def Gamma2(m, gamma0, q2, q02, L, m0, d):
    """
    Running width in the RBW

    .. math::
        \\Gamma(m) = \\Gamma_0 \\left(\\frac{q}{q_0}\\right)^{2L+1}\\frac{m_0}{m} B_{L}'^2(q,q_0,d)

    """
    q02 = tf.cast(q02, q2.dtype)
    _epsilon = 1e-15
    qq0 = q2 / q02
    qq0 = tf.cast(qq0 ** L, tf.complex128) * tf.sqrt(
        tf.cast(qq0, tf.complex128)
    )
    mm0 = tf.cast(m0, m.dtype) / m
    z0 = q02 * d ** 2
    z = q2 * d ** 2
    bp = Bprime_polynomial(L, z0) / Bprime_polynomial(L, z)
    gammaM = qq0 * tf.cast(gamma0 * bp * mm0, qq0.dtype)
    return gammaM


def Bprime_q2(L, q2, q02, d):
    """
    Blatt-Weisskopf barrier factors.
    """
    q02 = tf.cast(q02, q2.dtype)
    _epsilon = 1e-15
    z0 = q02 * d ** 2
    z = q2 * d ** 2
    bp = Bprime_polynomial(L, z0) / Bprime_polynomial(L, z)
    return tf.sqrt(tf.where(bp > 0, bp, 1.0))


def Bprime_num(L, q, d):
    """
    The numerator (as well as the denominator) inside the square root in the barrier factor

    """
    z = (q * d) ** 2
    bp = Bprime_polynomial(L, z)
    return tf.sqrt(bp)


def Bprime(L, q, q0, d):
    """
    Blatt-Weisskopf barrier factors. E.g. the first three orders

    ===========  ===================================================
     :math:`L`                :math:`B_L(q,q_0,d)`
    ===========  ===================================================
     0                                 1
     1                 :math:`\\sqrt{\\frac{(q_0d)^2+1}{(qd)^2+1}}`
     2            :math:`\\sqrt{\\frac{(q_0d)^4+3*(q_0d)^2+9}{(qd)^4+3*(qd)^2+9}}`
    ===========  ===================================================

    :math:`d` is 3.0 by default.
    """
    num = Bprime_num(L, q0, d)
    denom = Bprime_num(L, q, d)
    return tf.cast(num, denom.dtype) / denom


def barrier_factor(l, q, q0, d=3.0, axis=0):  # cache q^l * B_l 只用于H里
    """
    Barrier factor multiplied with :math:`q^L`, which is used as a combination in the amplitude expressions. The values
    are cached for :math:`L` ranging from 0 to **l**.
    """
    ret = []
    for i in l:
        tmp = q ** i * tf.cast(Bprime(i, q, q0, d), q.dtype)
        ret.append(tmp)
    return tf.stack(ret)


def barrier_factor2(l, q, q0, d=3.0, axis=-1):  # cache q^l * B_l 只用于H里
    """
    ???
    """
    ret = []
    for i in l:
        tmp = q ** i * tf.cast(Bprime(i, q, q0, d), q.dtype)
        ret.append(tf.reshape(tmp, (-1, 1)))
    return tf.concat(ret, axis=axis)


def Bprime_polynomial(l, z):
    """
    It stores the Blatt-Weisskopf polynomial up to the fifth order (:math:`L=5`)

    :param l: The order
    :param z: The variable in the polynomial
    :return: The calculated value
    """
    coeff = {
        0: [1.0],
        1: [1.0, 1.0],
        2: [1.0, 3.0, 9.0],
        3: [1.0, 6.0, 45.0, 225.0],
        4: [1.0, 10.0, 135.0, 1575.0, 11035.0],
        5: [1.0, 15.0, 315.0, 6300.0, 99225.0, 893025.0],
    }
    l = int(l + 0.01)
    if l not in coeff:
        coeff[l] = [float(i) for i in get_bprime_coeff(l)]
        # raise NotImplementedError
    z = tf.convert_to_tensor(z)
    cof = [tf.convert_to_tensor(i, z.dtype) for i in coeff[l]]
    ret = tf.math.polyval(cof, z)
    return ret


def reverse_bessel_polynomials(n, x):
    """Reverse Bessel polynomials.

    .. math::
        \\theta_{n}(x) = \\sum_{k=0}^{n} \\frac{(n+k)!}{(n-k)!k!} \\frac{x^{n-k}}{2^k}

    """
    ret = 0
    for k in range(n + 1):
        c = fractions.Fraction(
            math.factorial(n + k),
            math.factorial(n - k) * math.factorial(k) * 2 ** k,
        )
        ret += c * x ** (n - k)
    return ret


@functools.lru_cache()
def get_bprime_coeff(l):
    """The coefficients of polynomial in Bprime function.

    .. math::
        |\\theta_{l}(jw)|^2 = \\sum_{i=0}^{l} c_i w^{2 i}

    """
    x = sym.Symbol("x")
    theta = reverse_bessel_polynomials(l, x)
    w = sym.Symbol("w", real=True)
    Hjw = theta.subs({"x": sym.I * w})
    Hjw2 = sym.Poly(Hjw * Hjw.conjugate(), w)
    coeffs = Hjw2.as_dict()
    ret = [coeffs.get((2 * l - 2 * i,), 0) for i in range(l + 1)]
    return ret



import pytest

from tf_pwa.amp import *
from tf_pwa.cal_angle import cal_angle_from_momentum
from tf_pwa.model import FCN, Model

from .common import write_temp_file


@regist_particle("Gounaris–Sakurai")
class ParticleGS(Particle):
    def get_amp(self, data, data_c):
        r"""
        Gounaris G.J., Sakurai J.J., Phys. Rev. Lett., 21 (1968), pp. 244-247

        .. math::
          R(m) = \frac{1 + D \Gamma_0 / m_0}{(m_0^2 -m^2) + f(m) - i m_0 \Gamma(m)}

        .. math::
          f(m) = \Gamma_0 \frac{m_0 ^2 }{q_0^3} \left[q^2 [h(m)-h(m_0)] + (m_0^2 - m^2) q_0^2 \frac{d h}{d m}|_{m0} \right]

        .. math::
          h(m) = \frac{2}{\pi} \frac{q}{m} \ln \left(\frac{m+q}{2m_{\pi}} \right)

        .. math::
          \frac{d h}{d m}|_{m0} = h(m_0) [(8q_0^2)^{-1} - (2m_0^2)^{-1}] + (2\pi m_0^2)^{-1}

        .. math::
          D = \frac{f(0)}{\Gamma_0 m_0} = \frac{3}{\pi}\frac{m_\pi^2}{q_0^2} \ln \left(\frac{m_0 + 2q_0}{2 m_\pi }\right)
            + \frac{m_0}{2\pi q_0} - \frac{m_\pi^2 m_0}{\pi q_0^3}
        """
        raise NotImplementedError


def get_test_decay():
    a = Particle("A", J=1, P=-1, spins=(-1, 1))
    b = Particle("B", J=1, P=-1)
    c = Particle("C", J=0, P=-1)
    d = Particle("D", J=1, P=-1)
    bd = ParticleLass("BD", 1, 1, mass=1.0, width=1.0)
    cd = Particle("CD", 1, 1, mass=1.0, width=1.0)
    bc = Particle("BC", 1, 1, mass=1.0, width=1.0)
    R = get_particle("R", 1, 1, mass=1.0, width=1.0)
    HelicityDecay(a, [bc, d])
    HelicityDecay(bc, [b, c])
    HelicityDecay(a, [cd, b])
    HelicityDecay(cd, [c, d])
    get_decay(a, [bd, c], model="helicity_full-bf")
    get_decay(bd, [b, d])
    HelicityDecayNP(a, [R, c])
    HelicityDecayP(R, [b, d])
    d3 = AngSam3Decay(a, [b, c, d])
    de = DecayGroup(a.chain_decay())
    print(de)
    return de, [b, c, d]


test_data = [
    [
        np.array([[2.0, 0.1, 0.2, 0.3]]),
        np.array([[3.0, 0.2, 0.3, 0.4]]),
        np.array([[4.0, 0.3, 0.4, 0.5]]),
    ],
    [
        np.array([[2.0, 0.1, 0.2, 0.3], [2.0, 0.3, 0.2, 0.1]]),
        np.array([[3.0, 0.2, 0.3, 0.4], [3.0, 0.4, 0.3, 0.2]]),
        np.array([[4.0, 0.3, 0.4, 0.5], [4.0, 0.5, 0.4, 0.3]]),
    ],
]


def test_amp():
    decs, particle = get_test_decay()
    amp = AmplitudeModel(decs)
    for p_data in test_data:
        p = dict(zip(particle, p_data))
        data = cal_angle_from_momentum(p, decs)
        amp(data)


def test_valid_jp():
    a = get_particle("a", J=0.5, P=-1)
    b = get_particle("b", J=0, P=+1)
    c = get_particle("c", J=1, P=-1)
    d = get_particle("d", J=0, P=-1)
    print(get_decay(a, [c, d], p_break=True).get_ls_list())
    with pytest.raises(ValueError):
        get_decay(b, [c, d]).check_valid_jp()
    with pytest.raises(ValueError):
        get_decay(a, [c, d], p_break=True).check_valid_jp()
    get_decay(b, [c, d], p_break=True).check_valid_jp()


def test_get_relative_momentum():
    a = get_particle("a", J=0, P=1)
    b = get_particle("b", J=0, P=1)
    c = get_particle("c", J=0, P=1, mass=2.0)
    a.get_mass()
    d = get_decay(a, [b, c])
    d.init_params()
    data = {
        a: {"m": np.array([5.0])},
        b: {"m": np.array([1.0])},
        c: {"m": np.array([2.0])},
    }
    p1 = d.get_relative_momentum(data)
    p2 = d.get_relative_momentum(data, False)

    p3 = d.get_relative_momentum2(data)
    p4 = d.get_relative_momentum2(data, False)

    assert p1 ** 2 == p3
    assert p2 ** 2 == p4


def test_simple_resonances():
    @simple_resonance("xxx")
    def f(m, s=3.0):
        return m + s

    b = get_particle("ss:2", model="xxx")
    b.init_params()
    b(1.0, s=2.0)
    b(2.0)

    a = b.get_amp({"m": 1.0}, {"|q|": 1.0}, {})
    assert np.allclose(np.array(4.0 + 0.0j), a.numpy())

    @simple_resonance("xxx2")
    def g(m, m0, g0, q, q0, a: FloatParams = 2.0):
        return m + a + q + q0

    b = get_particle("ss:2", a=3.0, model="xxx2")
    b.init_params()
    a = b.get_amp({"m": 1.0}, {"|q|": 1.0, "|q0|": 1.0}, {})
    assert np.allclose(np.array(6.0 + 0.0j), a.numpy())


def test_flatte():
    a = get_particle(
        "flatte", model="Flatte", mass=3.6, mass_list=[[1.0, 1.0], [1.5, 2.0]]
    )
    a.init_params()
    b = a(np.array([3.6]))
    assert b.numpy().real == 0


def test_gs():
    a = get_particle("gs", J=1, P=-1, model="GS_rho", mass=3.6, width=0.01)
    b = [get_particle(i, J=0, P=-1) for i in "ac"]
    get_decay(a, b)
    a.init_params()
    b = a.get_amp(
        {"m": np.array(1.0)},
        {"|q|": np.array(1.0), "|q0|": np.array(1.0)},
        all_data={},
    )


def test_BW():
    a = get_particle("BW", J=1, P=-1, model="BW", mass=3.6, width=0.01)
    b = [get_particle(i, J=0, P=-1) for i in "ac"]
    get_decay(a, b)
    a.init_params()
    b = a.get_amp(
        {"m": np.array(1.0)},
        {"|q|": np.array(1.0), "|q0|": np.array(1.0)},
        all_data={},
    )


def simple_run(name):
    a = get_particle(name, J=1, P=-1, model=name, mass=3.6, width=0.01)
    b = [get_particle(i, J=0, P=-1) for i in "ac"]
    get_decay(a, b)
    a.init_params()
    b = a.get_amp(
        {"m": np.array(1.0)},
        {
            "|q|": np.array(1.0),
            "|q0|": np.array(1.0),
            "|q|2": np.array(1.0),
            "|q0|2": np.array(1.0),
        },
        all_data={},
    )


def test_one():
    simple_run("one")
    simple_run("exp")
    simple_run("exp_com")
    simple_run("BWR_normal")


def test_model_new():
    decs, particle = get_test_decay()
    amp = AmplitudeModel(decs)
    data = []
    for p_data in test_data:
        p = dict(zip(particle, p_data))
        data.append(cal_angle_from_momentum(p, decs))
    model = Model(amp)
    fcn = FCN(model, data[0], data[1])
    nll1, grad1 = fcn.nll_grad({})
    nll2 = fcn({})
    nll3, grad3, he = fcn.nll_grad_hessian({})


def test_particle():
    a = get_particle("ss", model="Gounaris–Sakurai")
    with pytest.raises(NotImplementedError):
        a.get_amp({}, {})


def test_trans_model():
    a = get_particle("a", model={"expr": "-1*x+2j"})
    b = a(1.0)
    assert b.numpy() == -1 + 2j
    c = get_particle(
        "c", mass=2.0, width=2.0, model={"expr": "a*BW", "where": {"a": -1}}
    )
    d = c.get_amp({"m": np.array([1.0, 2.0])})
    assert np.allclose(d.numpy(), np.array([-0.12 - 0.16j, -0.0 - 0.25j]))
    e = get_particle("e", mass=2.0, width=2.0, model={"expr": "-BW"})
    f = e.get_amp({"m": np.array([1.0, 2.0])})
    assert np.allclose(f.numpy(), d.numpy())


def test_decay_params():
    a = get_particle(
        "a",
        decay_params={"a": 3},
    )
    b = get_particle(
        "b",
        production_params={"a": 1, "b": 2},
    )
    c = get_particle(
        "c",
        production_params={"a": 1, "c": 1},
    )
    dec = get_decay(a, [b, c])
    assert dec.a == 3
    assert dec.b == 2
    assert dec.c == 1


def test_decay_ls_amp():
    a = get_particle("a", mass=1.0, width=0.5, P=1)
    c = get_particle("c")
    d = get_particle("d")
    dec1 = get_decay(a, [c, d])
    dec1.init_params()
    rnd = np.random.random((10,))
    ma = np.ones_like(rnd)
    mb = 0.3 * np.ones_like(rnd)
    ang = {"alpha": rnd, "beta": rnd, "gamma": rnd}
    data = {c: ang, d: ang}
    data_p = {a: {"m": ma}, c: {"m": mb}, d: {"m": mb}}
    dec1.get_ls_amp(data, data_p)
    dec1.get_ls_amp({**data, "|q|2": mb}, data_p)


def test_polarization():
    a = get_particle("a", J=0.5, P=1, polarization="vector")
    c = get_particle("c", J=0.5, P=1)
    d = get_particle("d", J=0, P=-1)
    b = get_particle("b", J=0, P=-1)
    r = get_particle("r", mass=1.0, width=0.5, J=0.5, P=1)
    dec2 = get_decay(r, [c, d])
    dec1 = get_decay(a, [r, b])
    decs = DecayGroup([[dec1, dec2]])
    amp = AmplitudeModel(decs, polar=False)
    print(amp.get_params())
    for p_data in test_data:
        p = dict(zip([b, c, d], p_data))
        data = cal_angle_from_momentum(p, decs)
        amp(data)


def test_no_mass0():
    a = get_particle("a", mass=1.0, width=0.5, P=1)
    c = get_particle("c")
    d = get_particle("d")
    dec1 = get_decay(a, [c, d])
    decs = DecayGroup([[dec1]])
    amp = AmplitudeModel(decs, polar=False)
    assert all(not i for i in amp.vm.complex_vars.values())
    amp2 = AmplitudeModel(decs, polar=True)
    assert all(amp2.vm.complex_vars.values())


def test_dec():
    s = """

Particle D_2*+         2.4654          0.0467
Particle anti-D_2*0    2.4607          0.0475
RUNNINGWIDTH D_2*+

Decay vpho
1 anti-D_2*0  D*0    HELCOV gg6a aa6a gg7a aa7a gg8a aa8a gg9a aa9a  one zero;
1 D_2*+  D*-         HELCOV gg6a aa6a gg7a aa7a gg8a aa8a gg9a aa9a  one zero;
Enddecay

Decay anti-D_2*0
1  D*- pi+ HELCOV one zero;
Enddecay

Decay D_2*+
1  D*0 pi+ HELCOV one zero;
Enddecay

End
    """

    with write_temp_file(s) as f:
        top, inner, final = load_decfile_particle(f)

    assert top == {Particle("vpho")}
    assert final == {Particle(i) for i in ["D*0", "D*-", "pi+"]}
    assert inner == {Particle(i) for i in ["D_2*+", "anti-D_2*0"]}

    inner = list(inner)
    assert inner[0].decay[0].params == ["one", "zero"]


def test_split_ls():
    a = get_particle("a", J=1, P=1, mass=6.0, width=0.02, model="BWR_LS")
    b = get_particle("b", J=1, P=-1)
    c = get_particle("c", J=0, P=-1)
    d = get_particle("d", J=0, P=-1)
    t = get_particle("t", J=0, P=-1)
    dec = get_decay(a, [b, c])
    dec2 = get_decay(t, [a, d], p_break=True)
    dg = DecayGroup([DecayChain([dec, dec2])])
    amp = AmplitudeModel(dg)
    p = dict(zip([b, c, d], test_data[0]))
    data = cal_angle_from_momentum(p, dg)
    amp1 = amp(data)
    a(np.array([5.9, 6.0, 6.1]))



import yaml

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.config_loader.sample import build_phsp_chain_sorted
from tf_pwa.data import data_shape
from tf_pwa.tests.test_full import gen_toy, toy_config


def test_generate_phsp(toy_config):
    data = toy_config.generate_toy(1000, force=False)
    assert data_shape(data) >= 1000
    data = toy_config.generate_toy(1000)
    assert data_shape(data) == 1000
    data = toy_config.generate_toy2(1000)
    assert data_shape(data) == 1000
    data = toy_config.generate_toy2(1000, gen=toy_config.generate_phsp)
    assert data_shape(data) == 1000
    data = toy_config.generate_toy2(
        1000, gen_p=toy_config.generate_phsp_p, include_charge=True
    )
    data.savetxt(
        "toy_data/data_c.dat",
        toy_config.get_dat_order(),
        cp_trans=True,
        save_charge=True,
    )
    assert data_shape(data) == 1000
    data = toy_config.generate_toy2(1000, gen_p=toy_config.generate_phsp_p)
    assert data_shape(data) == 1000
    gen_p2 = lambda N: {
        str(k): v for k, v in toy_config.generate_phsp_p(N).items()
    }
    data = toy_config.generate_toy2(1000, gen_p=gen_p2)
    assert data_shape(data) == 1000


config_text = """
decay:
    A: [[R1, B], [R2, C]]
    R1: [C, D]
    R2: [B, D]
    D: [E, F]
    E: [G, H]

particle:
    $top:
        A: {m0: 2.0}
    $finals:
        B: {m0: 0.3}
        C: {m0: 0.3}
        F: {m0: 0.3}
        G: {m0: 0.1}
        H: {m0: 0.1}
    E: {m0: 0.3, J: 0, P: 1, model: one}
    R1: {m0: 1.4, g0: 0.05, J: 0, P: 1}
    R2: {m0: 1.3, g0: 0.03, J: 0, P: 1}
    D: {m0: 0.8, model: one}

"""


def test_chain_phsp():
    dic = yaml.full_load(config_text)
    config = ConfigLoader(dic)
    data = config.generate_toy(10)
    data.mass_hist("(C, F, G, H)").draw()
    import matplotlib.pyplot as plt

    plt.savefig("chain_phsp.png")



from tf_pwa.cg import cg_coef, get_cg_coef
from tf_pwa.significance import significance


def close_to(x, y, err=1e-5):
    return abs(x - y) < err


def test_significance():
    args = [
        [0.5, 0.0, 1],
        [0.0, 0.5, 1],
        [1.0, 0.5, 1],
        [-2.0, -3.0, 2],
        [-4.0, -4.0, 5],
    ]
    results = [1.0, 1.0, 1.0, 0.9004525966377901, 0.0]
    for i, j in zip(args, results):
        assert close_to(significance(i[0], i[1], i[2]), j)


def test_get_cg_coef():
    assert close_to(get_cg_coef(1, 1, 1, 0, 2, 1), cg_coef(1, 1, 1, 0, 2, 1))
    assert close_to(
        get_cg_coef(2, 1, -1, 0, 2, -1), cg_coef(2, 1, -1, 0, 2, -1)
    )
    assert close_to(get_cg_coef(2, 1, -1, 1, 2, 0), cg_coef(2, 1, -1, 1, 2, 0))



import numpy as np

from tf_pwa.cal_angle import *
from tf_pwa.data import *


def test_process():
    a, b, c, d = [BaseParticle(i) for i in ["A", "B", "C", "D"]]
    p = {
        b: np.array([[1.0, 0.2, 0.3, 0.2]]),
        c: np.array([[2.0, 0.1, 0.3, 0.4]]),
        d: np.array([[3.0, 0.2, 0.5, 0.7]]),
    }
    # st = {b: [b], c: [c], d: [d], a: [b, c, d], r: [b, d]}
    decs = DecayGroup(DecayChain.from_particles(a, [b, c, d]))
    print(decs)
    data = cal_angle_from_momentum(p, decs)
    assert isinstance(data, CalAngleData)
    data.get_weight()
    data = add_weight(data)
    print(data_shape(data, all_list=True))
    print(len(list(split_generator(data, 5000))))
    data = data_to_numpy(data)
    assert data_shape(data) == 1
    data.get_weight()
    data.get_mass("(B, C)")
    dec = data.get_decay().get_decay_chain("(B, C)")
    dec2 = data.get_decay()[0]
    ang = data.get_angle(dec, "B")
    ang2 = data.get_angle("(B, C)", "B")
    assert ang is ang2
    assert "alpha" in ang
    assert "beta" in ang
    assert np.allclose(ang["gamma"], 0)
    p = data.get_momentum("(B, D)")
    assert p.shape[-1] == 4

    data.savetxt("cal_angle_file.txt", ["C", "D"])
    data.savetxt("cal_angle_file.txt")
    hist = data.mass_hist("(C, D)")



#!/usr/bin/env python3

import json
import math
import os.path
import sys
import time

import iminuit
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from fit_scipy import prepare_data
from scipy.optimize import BFGS, basinhopping, minimize

from tf_pwa.angle import EularAngle, cal_ang_file
from tf_pwa.bounds import Bounds
from tf_pwa.model import FCN, Cache_Model, param_list
from tf_pwa.utils import flatten_np_data, load_config_file, pprint

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def main(param_name, x, method):
    dtype = "float64"
    w_bkg = 0.768331
    # set_gpu_mem_growth()
    tf.keras.backend.set_floatx(dtype)

    config_list = load_config_file("Resonances")
    data, bg, mcdata = prepare_data(dtype=dtype)
    a = Cache_Model(config_list, w_bkg, data, mcdata, bg=bg, batch=65000)
    try:
        with open("lklpf_params.json") as f:
            param = json.load(f)
            a.set_params(param["value"])
    except:
        pass
    pprint(a.get_params())

    fcn = FCN(a)  # 1356*18

    def LP_minuit(param_name, fixed_var):
        args = {}
        args_name = []
        x0 = []
        bounds_dict = {
            param_name: (fixed_var, fixed_var),
            "Zc_4160_m0:0": (4.1, 4.22),
            "Zc_4160_g0:0": (0, 10),
        }
        for i in a.Amp.trainable_variables:
            args[i.name] = i.numpy()
            x0.append(i.numpy())
            args_name.append(i.name)
            args["error_" + i.name] = 0.1
            if i.name not in bounds_dict:
                bounds_dict[i.name] = (0.0, None)
        for i in bounds_dict:
            if i in args_name:
                args["limit_{}".format(i)] = bounds_dict[i]
        m = iminuit.Minuit(
            fcn,
            forced_parameters=args_name,
            errordef=0.5,
            grad=fcn.grad,
            print_level=2,
            use_array_call=True,
            **args,
        )
        now = time.time()
        with tf.device("/device:GPU:0"):
            print(m.migrad(ncall=10000))  # ,precision=5e-7))
        print(time.time() - now)
        print(m.get_param_states())
        return m

    def LP_sp(param_name, fixed_var):
        args = {}
        args_name = []
        x0 = []
        bnds = []
        bounds_dict = {
            param_name: (fixed_var, fixed_var),
            "Zc_4160_m0:0": (4.1, 4.22),
            "Zc_4160_g0:0": (0, None),
        }
        for i in a.Amp.trainable_variables:
            args[i.name] = i.numpy()
            x0.append(i.numpy())
            args_name.append(i.name)
            if i.name in bounds_dict:
                bnds.append(bounds_dict[i.name])
            else:
                bnds.append((None, None))
            args["error_" + i.name] = 0.1
        now = time.time()
        bd = Bounds(bnds)
        f_g = bd.trans_f_g(fcn.nll_grad)
        callback = lambda x: print(fcn.cached_nll)
        with tf.device("/device:GPU:0"):
            # s = basinhopping(f.nll_grad,np.array(x0),niter=6,disp=True,minimizer_kwargs={"jac":True,"options":{"disp":True}})
            # 优化器
            # s = minimize(fcn.nll_grad,np.array(x0),method="L-BFGS-B",jac=True,bounds=bnds,callback=callback,options={"disp":1,"maxcor":100})
            s = minimize(
                f_g,
                np.array(bd.get_x(x0)),
                method="BFGS",
                jac=True,
                callback=callback,
                options={"disp": 1},
            )
        print("#" * 5, param_name, fixed_var, "#" * 5)
        # print(s)
        return s

    # x=np.arange(0.51,0.52,0.01)
    y = []
    if method == "scipy":
        for v in x:
            y.append(LP_sp(param_name, v).fun)
    elif method == "iminuit":
        for v in x:
            y.append(LP_minuit(param_name, v).get_fmin().fval)
    # print("lklhdx",x)
    # print("lklhdy",y)
    print("\nend\n")
    return y


def lklpf(param_name):
    with open("lklpf_params.json") as f:
        params = json.load(f)
    x_mean = params["value"][param_name]
    x_sigma = params["error"][param_name]
    method = "scipy"  ###
    mode = "bothsides"  # "back&forth"
    if mode == "back&forth":
        x1 = np.arange(
            x_mean - 5 * x_sigma, x_mean + 5 * x_sigma, x_sigma / 2
        )  ###
        x2 = x1[::-1]
        t1 = time.time()
        y1 = main(param_name, x1, method)
        t2 = time.time()
        y2 = main(param_name, x2, method)
        t3 = time.time()
        print(mode, x1, y1, x1, y2[::-1], sep="\n")
    elif mode == "bothsides":
        # x1=np.arange(x_mean,x_mean-5*x_sigma,-x_sigma/2)
        x1 = np.arange(x_mean, x_mean - 100, -10)
        # x2=np.arange(x_mean,x_mean+5*x_sigma,x_sigma/2)
        x2 = np.arange(x_mean, x_mean + 100, 10)
        t1 = time.time()
        y1 = main(param_name, x1, method)
        t2 = time.time()
        y2 = main(param_name, x2, method)
        t3 = time.time()
        print(
            mode,
            list(np.append(x1[::-1], x2)),
            list(np.append(y1[::-1], y2)),
            sep="\n",
        )
    print(param_name, x_mean)
    print("#" * 10, t2 - t1, "#" * 10, t3 - t2)
    """plt.plot(x,yf,"*-",x,yb,"*-")
  plt.title(param_name)
  plt.legend(["forward","backward"])
  plt.savefig("lklhd_prfl")
  plt.clf()"""


if __name__ == "__main__":
    param_list = ["D2_2460_BLS_2_1r:0"]
    for param_name in param_list:
        lklpf(param_name)
    print("\n*** likelihood profile done *****\n")



"""
This module provides functions which are useful when calculating the angular variables.


The full data structure provided is ::

    {
        "particle": {
            A: {"p": ..., "m": ...},
            (C, D): {"p": ..., "m": ...},
            ...
        },

        "decay":{
            [A->(C, D)+B, (C, D)->C+D]:
            {
                A->(C, D)+B: {
                    (C, D): {
                        "ang": {
                            "alpha":[...],
                            "beta": [...],
                            "gamma": [...]
                        },
                        "z": [[x1,y1,z1],...],
                        "x": [[x2,y2,z2],...]
                    },
                    B: {...}
                },

                (C, D)->C+D: {
                    C: {
                        ...,
                        "aligned_angle": {
                        "alpha": [...],
                        "beta": [...],
                        "gamma": [...]
                    }
                },
                    D: {...}
                },
                A->(B, D)+C: {...},
                (B, D)->B+D: {...}
            },
        ...
        }
    }


Inner nodes are named as tuple of particles.

"""
import itertools

import numpy as np

from .angle import SU2M, EulerAngle, LorentzVector, Vector3, _epsilon
from .config import get_config
from .data import (
    data_index,
    data_merge,
    data_shape,
    data_strip,
    data_to_numpy,
    flatten_dict_data,
    load_dat_file,
    split_generator,
)
from .histogram import Hist1D
from .particle import BaseDecay, BaseParticle, DecayChain, DecayGroup
from .tensorflow_wrapper import tf


class CalAngleData(dict):
    def get_decay(self):
        return DecayGroup(list(self["decay"].keys()))

    def get_mass(self, name):
        return data_index(self, ("particle", name, "m"))

    def get_momentum(self, name):
        return data_index(self, ("particle", name, "p"))

    def get_weight(self):
        if "weight" in self:
            return self["weight"]
        return tf.ones(data_shape(self), dtype=get_config("dtype"))

    def get_angle(self, decay, p):
        """ get hilicity angle of decay which product particle p"""
        if isinstance(decay, str):
            decay = self.get_decay().get_decay_chain(decay)
        dec = decay.standard_topology()
        dec_map = decay.topology_map()
        dec_i = decay[0]
        for i in decay:
            if str(p) in [str(j) for j in i.outs]:
                dec_i = i
                break
        p_name = data_index(dec_map, p)
        dec_name = dec_map[dec_i]
        return data_index(self, ("decay", dec, dec_name, p_name, "ang"))

    def mass_hist(self, name, bins="sqrt", **kwargs):
        data = data_to_numpy(self.get_mass(name))
        return Hist1D.histogram(data, bins=bins, **kwargs)

    def savetxt(
        self, file_name, order=None, cp_trans=False, save_charge=False
    ):
        if order is None:
            order = self.get_decay().outs
        pi = [data_to_numpy(self.get_momentum(i)) for i in order]
        if cp_trans:
            pi = [[i[0], -i[1], -i[2], -i[3]] for i in pi]
        pi = np.stack(pi).transpose((1, 0, 2)).reshape((-1, 4))
        np.savetxt(file_name, pi)
        if save_charge:
            np.savetxt(
                file_name[::-1].replace(".", ".c", 1)[::-1],
                self["charge_conjugation"],
            )


def struct_momentum(p, center_mass=True) -> dict:
    """
    restructure momentum as dict
        {outs:momentum} => {outs:{p:momentum}}
    """
    ret = {}
    if center_mass:
        ps_top = []
        for i in p:
            ps_top.append(p[i])
        p_top = tf.reduce_sum(ps_top, 0)
        for i in p:
            ret[i] = {"p": LorentzVector.rest_vector(p_top, p[i])}
    else:
        for i in p:
            ret[i] = {"p": p[i]}
    return ret


# data process
def infer_momentum(data, decay_chain: DecayChain) -> dict:
    """
    infer momentum of all particles in the decay chain from outer particles momentum.
        {outs:{p:momentum}} => {top:{p:momentum},inner:{p:..},outs:{p:..}}
    """
    st = decay_chain.sorted_table()
    for i in st:
        if i in data:
            continue
        ps = []
        for j in st[i]:
            ps.append(data[j]["p"])
        data[i] = {"p": tf.reduce_sum(ps, 0)}
    return data


def add_mass(data: dict, _decay_chain: DecayChain = None) -> dict:
    """
    add particles mass array for data momentum.
        {top:{p:momentum},inner:{p:..},outs:{p:..}} => {top:{p:momentum,m:mass},...}
    """
    for i in data:
        if isinstance(i, BaseParticle):
            p = data[i]["p"]
            data[i]["m"] = LorentzVector.M(p)
    return data


def add_weight(data: dict, weight: float = 1.0) -> dict:
    """
    add inner data weights for data.
        {...} => {..., "weight": weights}
    """
    data_size = data_shape(data)
    weight = [weight] * data_size
    data["weight"] = np.array(weight)
    return data


def cal_chain_boost(data, decay_chain: DecayChain) -> dict:
    """
    calculate chain boost for a decay chain
    """
    part_data = {}
    core_decay_map = {}

    decay_set = list(decay_chain)
    particle_set = set(decay_chain.inner) | set(decay_chain.outs)
    while len(decay_set) > 0:
        tmp_decay_set = []
        for i in decay_set:
            if i.core == decay_chain.top:
                part_data[i] = {}
                p_rest = data[i.core]["p"]
                part_data[i]["rest_p"] = {}
                for j in i.outs:
                    core_decay_map[j] = i
                    pj = data[j]["p"]
                    p = LorentzVector.rest_vector(p_rest, pj)
                    part_data[i]["rest_p"][j] = p
                    particle_set.remove(j)
                for j in particle_set:
                    pj = data[j]["p"]
                    p = LorentzVector.rest_vector(p_rest, pj)
                    part_data[i]["rest_p"][j] = p
            elif i.core in core_decay_map:
                part_data[i] = {}
                p_rest = part_data[core_decay_map[i.core]]["rest_p"][i.core]
                part_data[i]["rest_p"] = {}
                for j in i.outs:
                    core_decay_map[j] = i
                    pj = part_data[core_decay_map[i.core]]["rest_p"][j]
                    p = LorentzVector.rest_vector(p_rest, pj)
                    part_data[i]["rest_p"][j] = p
                    particle_set.remove(j)
                for j in particle_set:
                    pj = part_data[core_decay_map[i.core]]["rest_p"][j]
                    p = LorentzVector.rest_vector(p_rest, pj)
                    part_data[i]["rest_p"][j] = p
            else:
                tmp_decay_set.append(i)
        decay_set = tmp_decay_set
    # from pprint import pprint
    # pprint(part_data)
    # exit()
    return part_data


def cal_single_boost(data, decay_chain: DecayChain) -> dict:
    part_data = {}
    # part_data = cal_chain_boost(data, decay_chain)
    for i in decay_chain:
        part_data[i] = {}
        p_rest = data[i.core]["p"]
        part_data[i]["rest_p"] = {}
        for j in i.outs:
            pj = data[j]["p"]
            p = LorentzVector.rest_vector(p_rest, pj)
            part_data[i]["rest_p"][j] = p
    return part_data


# from pysnooper import snoop
# @snoop()
def cal_helicity_angle(
    data: dict,
    decay_chain: DecayChain,
    base_z=np.array([[0.0, 0.0, 1.0]]),
    base_x=np.array([[1.0, 0.0, 0.0]]),
) -> dict:
    """
    Calculate helicity angle for A -> B + C: :math:`\\theta_{B}^{A}, \\phi_{B}^{A}` from momentum.

    from `{A:{p:momentum},B:{p:...},C:{p:...}}`

    to   `{A->B+C:{B:{"ang":{"alpha":...,"beta":...,"gamma":...},"x":...,"z"},...}}`
    """
    ret = {}
    # boost all in them mother rest frame

    # print(decay_chain, part_data)
    part_data = cal_chain_boost(data, decay_chain)
    # print(decay_chain , part_data)
    # calculate angle and base x,z axis from mother particle rest frame momentum and base axis
    set_x = {decay_chain.top: base_x}
    set_z = {decay_chain.top: base_z}
    r_matrix = {}
    b_matrix = {}
    set_decay = list(decay_chain)
    while set_decay:
        extra_decay = []
        for i in set_decay:
            if i.core in set_x:
                ret[i] = {}
                bias = -np.pi
                for j in i.outs:
                    ret[i][j] = {}
                    p_rest = part_data[i]["rest_p"][j]
                    z2 = LorentzVector.vect(p_rest)
                    ang, x = EulerAngle.angle_zx_z_getx(
                        set_z[i.core], set_x[i.core], z2
                    )
                    set_x[j] = x
                    set_z[j] = z2
                    # set range to make sure opposite allow be - phi
                    ang["alpha"] = (ang["alpha"] - bias) % (2 * np.pi) + bias
                    bias -= np.pi
                    ret[i][j]["ang"] = ang
                    ret[i][j]["x"] = x
                    ret[i][j]["z"] = z2
                    Bp = SU2M.Boost_z_from_p(p_rest)
                    b_matrix[j] = Bp
                    r = SU2M.Rotation_y(ang["beta"]) * SU2M.Rotation_z(
                        ang["alpha"]
                    )
                    if i.core in r_matrix:
                        r_matrix[j] = r * b_matrix[i.core] * r_matrix[i.core]
                    else:
                        r_matrix[j] = r
                if len(i.outs) == 3:
                    # Euler Angle for
                    p_rest = [part_data[i]["rest_p"][j] for j in i.outs]
                    zi = [LorentzVector.vect(i) for i in p_rest]
                    ret[i]["ang"], xi = EulerAngle.angle_zx_zzz_getx(
                        set_z[i.core], set_x[i.core], zi
                    )
                    for j, x, z, p_rest_i in zip(i.outs, xi, zi, p_rest):
                        ret[i][j] = {}
                        ret[i][j]["x"] = x
                        ret[i][j]["z"] = z
                        Bp = SU2M.Boost_z_from_p(p_rest_i)
                        b_matrix[j] = Bp
                        r = SU2M.Rotation_y(ang["beta"]) * SU2M.Rotation_z(
                            ang["alpha"]
                        )
                        if i.core in r_matrix:
                            r_matrix[j] = (
                                r * b_matrix[i.core] * r_matrix[i.core]
                            )
                        else:
                            r_matrix[j] = r
            else:
                extra_decay.append(i)
        set_decay = extra_decay
    ret["r_matrix"] = r_matrix
    ret["b_matrix"] = b_matrix
    return ret


def cal_angle_from_particle(
    data,
    decay_group: DecayGroup,
    using_topology=True,
    random_z=True,
    r_boost=True,
    final_rest=True,
):
    """
    Calculate helicity angle for particle momentum, add aligned angle.

    :params data: dict as {particle: {"p":...}}

    :return: Dictionary of data
    """
    if using_topology:
        decay_chain_struct = decay_group.topology_structure()
    else:
        decay_chain_struct = decay_group
    decay_data = {}

    # get base z axis
    p4 = data[decay_group.top]["p"]
    p3 = LorentzVector.vect(p4)
    base_z = np.array([[0.0, 0.0, 1.0]]) + tf.zeros_like(p3)
    if random_z:
        p3_norm = Vector3.norm(p3)
        mask = tf.expand_dims(p3_norm < 1e-5, -1)
        base_z = tf.where(mask, base_z, p3)
    # calculate chain angle
    for i in decay_chain_struct:
        data_i = cal_helicity_angle(data, i, base_z=base_z)
        decay_data[i] = data_i

    # calculate aligned angle of final particles in each decay chain
    set_x = {}  # reference particles
    ref_matrix = {}
    # for particle from a the top rest frame
    for idx, decay_chain in enumerate(decay_chain_struct):
        for decay in decay_chain:
            if decay.core == decay_group.top:
                for i in decay.outs:
                    if (i not in set_x) and (i in decay_group.outs):
                        set_x[i] = (decay_chain, decay)
                        ref_matrix[i] = decay_chain
    # or in the first chain
    for i in decay_group.outs:
        if i not in set_x:
            decay_chain = next(iter(decay_chain_struct))
            for decay in decay_chain:
                for j in decay.outs:
                    if i == j:
                        set_x[i] = (decay_chain, decay)
                        ref_matrix[i] = decay_chain
    for idx, decay_chain in enumerate(decay_chain_struct):
        for decay in decay_chain:
            part_data = decay_data[decay_chain][decay]
            for i in decay.outs:
                if i in decay_group.outs and decay_chain != set_x[i][0]:
                    idx2, decay2 = set_x[i]
                    part_data2 = decay_data[idx2][decay2]
                    if r_boost:
                        r_matrix = decay_data[decay_chain]["r_matrix"][i]
                        b_matrix = decay_data[decay_chain]["b_matrix"][i]
                        r_matrix_ref = decay_data[ref_matrix[i]]["r_matrix"][i]
                        b_matrix_ref = decay_data[ref_matrix[i]]["b_matrix"][i]
                        R = SU2M(r_matrix_ref["x"]) * SU2M.inv(r_matrix)
                        # print(R)
                        if final_rest:
                            R = (
                                SU2M(b_matrix_ref["x"])
                                * R
                                * SU2M.inv(b_matrix)
                            )
                        ang = R.get_euler_angle()
                    else:
                        x1 = part_data[i]["x"]
                        x2 = part_data2[i]["x"]
                        z1 = part_data[i]["z"]
                        z2 = part_data2[i]["z"]
                        ang = EulerAngle.angle_zx_zx(z1, x1, z2, x2)
                    # ang = AlignmentAngle.angle_px_px(z1, x1, z2, x2)
                    part_data[i]["aligned_angle"] = ang
    ret = data_strip(decay_data, ["r_matrix", "b_matrix", "x", "z"])
    return ret


cal_angle = cal_angle_from_particle


def Getp(M_0, M_1, M_2):
    """
    Consider a two-body decay :math:`M_0\\rightarrow M_1M_2`. In the rest frame of :math:`M_0`, the momentum of
    :math:`M_1` and :math:`M_2` are definite.

    :param M_0: The invariant mass of :math:`M_0`
    :param M_1: The invariant mass of :math:`M_1`
    :param M_2: The invariant mass of :math:`M_2`
    :return: the momentum of :math:`M_1` (or :math:`M_2`)
    """
    M12S = M_1 + M_2
    M12D = M_1 - M_2
    p = (M_0 - M12S) * (M_0 + M12S) * (M_0 - M12D) * (M_0 + M12D)
    q = (
        p + tf.abs(p)
    ) / 2  # if p is negative, which results from bad data, the return value is 0.0
    return tf.sqrt(q) / (2 * M_0)


def Getp2(M_0, M_1, M_2):
    """
    Consider a two-body decay :math:`M_0\\rightarrow M_1M_2`. In the rest frame of :math:`M_0`, the momentum of
    :math:`M_1` and :math:`M_2` are definite.

    :param M_0: The invariant mass of :math:`M_0`
    :param M_1: The invariant mass of :math:`M_1`
    :param M_2: The invariant mass of :math:`M_2`
    :return: the momentum of :math:`M_1` (or :math:`M_2`)
    """
    M12S = M_1 + M_2
    M12D = M_1 - M_2
    p = (M_0 - M12S) * (M_0 + M12S) * (M_0 - M12D) * (M_0 + M12D)
    return p / (4 * M_0 * M_0)


def get_relative_momentum(data: dict, decay_chain: DecayChain):
    """
    add add rest frame momentum scalar from data momentum.

    from `{"particle": {A: {"m": ...}, ...}, "decay": {A->B+C: {...}, ...}`

    to `{"particle": {A: {"m": ...}, ...},"decay": {A->B+C:{...,"|q|": ...},...}`
    """
    ret = {}
    for decay in decay_chain:
        m0 = data[decay.core]["m"]
        m1 = data[decay.outs[0]]["m"]
        m2 = data[decay.outs[1]]["m"]
        p = Getp(m0, m1, m2)
        ret["decay"][decay] = {}
        ret["decay"][decay]["|q|"] = p
    return ret


def add_relative_momentum(data: dict):
    """
    add add rest frame momentum scalar from data momentum.

    from `{"particle": {A: {"m": ...}, ...}, "decay": {A->B+C: {...}, ...}`

    to `{"particle": {A: {"m": ...}, ...},"decay": {[A->B+C,...]: {A->B+C:{...,"|q|": ...},...},...}`
    """
    data_p = data["particle"]
    for decay_chain in data["decay"]:
        for decay in decay_chain:
            m0 = data_p[decay.core]["m"]
            m1 = data_p[decay.outs[0]]["m"]
            m2 = data_p[decay.outs[1]]["m"]
            p2 = Getp2(m0, m1, m2)
            data["decay"][decay_chain][decay]["|q|2"] = p2
    return data


def parity_trans(p, charges):
    charges = charges[: p.shape[0], None]
    return tf.where(charges > 0, p, LorentzVector.neg(p))


def prepare_data_from_decay(
    fnames, decs, particles=None, dtype=None, charges=None, **kwargs
):
    """
    Transform 4-momentum data in files for the amplitude model automatically via DecayGroup.

    :param fnames: File name(s).
    :param decs: DecayGroup
    :param particles: List of Particle. The final particles.
    :param dtype: Data type.
    :return: Dictionary
    """
    if dtype is None:
        dtype = get_config("dtype")
    if particles is None:
        particles = sorted(decs.outs)
    p = load_dat_file(fnames, particles, dtype=dtype)
    if charges is not None:
        p = {k: parity_trans(v, charges) for k, v in p.items()}
    data = cal_angle_from_momentum(p, decs, **kwargs)
    return data


def prepare_data_from_dat_file(fnames):
    """
    [deprecated] angle for amplitude.py
    """
    a, b, c, d = [BaseParticle(i) for i in ["A", "B", "C", "D"]]
    bc, cd, bd = [BaseParticle(i) for i in ["BC", "CD", "BD"]]
    p = load_dat_file(fnames, [d, b, c])
    # st = {b: [b], c: [c], d: [d], a: [b, c, d], r: [b, d]}
    decs = DecayGroup(
        [
            [BaseDecay(a, [bc, d]), BaseDecay(bc, [b, c])],
            [BaseDecay(a, [bd, c]), BaseDecay(bd, [b, d])],
            [BaseDecay(a, [cd, b]), BaseDecay(cd, [c, d])],
        ]
    )
    # decs = DecayChain.from_particles(a, [d, b, c])
    data = cal_angle_from_momentum(p, decs)
    data = data_to_numpy(data)
    data = flatten_dict_data(data)
    return data


def get_chain_data(data, decay_chain=None):
    """
    get all independent data for a decay chain
    """
    if decay_chain is None:
        decay_chain = list(data["decay"].keys())[0]
    chain_data = data["decay"][decay_chain]
    ret = {"mass": {}, "costheta": {}, "phi": {}}
    for dec in chain_data.keys():
        ret["mass"][dec.core] = data["particle"][dec.core]["m"]
        out1 = dec.outs[0]
        ang = chain_data[dec][out1]["ang"]
        ret["costheta"][dec] = tf.cos(ang["beta"])
        ret["phi"][dec] = ang["alpha"]
    return ret


def cal_angle_from_momentum_base(
    p,
    decs: DecayGroup,
    using_topology=True,
    center_mass=False,
    r_boost=True,
    random_z=False,
    batch=65000,
) -> CalAngleData:
    """
    Transform 4-momentum data in files for the amplitude model automatically via DecayGroup.

    :param p: 4-momentum data
    :param decs: DecayGroup
    :return: Dictionary of data
    """
    if data_shape(p) is None:
        return cal_angle_from_momentum_single(
            p, decs, using_topology, center_mass, r_boost, random_z
        )
    ret = []
    for i in split_generator(p, batch):
        ret.append(
            cal_angle_from_momentum_single(
                i, decs, using_topology, center_mass, r_boost, random_z
            )
        )
    return data_merge(*ret)


def identical_particles_swap(id_particles):
    ret = []
    for i in id_particles:
        ret.append(list(itertools.permutations(i)))
    for i in itertools.product(*ret):
        yield i


def identical_particles_swap_p(p4, id_particles):
    for comb in identical_particles_swap(id_particles):
        all_keys = tuple(p4.keys())
        name_map = {str(k): k for k in all_keys}
        swap_map = {}
        for c, p_list in zip(comb, id_particles):
            for ci, pi in zip(c, p_list):
                swap_map[ci] = name_map[pi]
        new_order = tuple([swap_map.get(str(i), i) for i in all_keys])
        if all_keys == new_order:
            continue
        yield (new_order, comb), dict(zip(new_order, p4.values()))


def cal_angle_from_momentum(
    p,
    decs: DecayGroup,
    using_topology=True,
    center_mass=False,
    r_boost=True,
    random_z=False,
    batch=65000,
) -> CalAngleData:
    """
    Transform 4-momentum data in files for the amplitude model automatically via DecayGroup.

    :param p: 4-momentum data
    :param decs: DecayGroup
    :return: Dictionary of data
    """
    ret = []
    id_particles = decs.identical_particles
    data = cal_angle_from_momentum_base(
        p, decs, using_topology, center_mass, r_boost, random_z, batch
    )
    if id_particles is None or len(id_particles) == 0:
        return data
    else:
        data["id_swap"] = {}
        for i, pi in identical_particles_swap_p(p, id_particles):
            data["id_swap"][i] = cal_angle_from_momentum_base(
                pi, decs, using_topology, center_mass, r_boost, random_z, batch
            )
        return data


def cal_angle_from_momentum_single(
    p,
    decs: DecayGroup,
    using_topology=True,
    center_mass=False,
    r_boost=True,
    random_z=True,
) -> CalAngleData:
    """
    Transform 4-momentum data in files for the amplitude model automatically via DecayGroup.

    :param p: 4-momentum data
    :param decs: DecayGroup
    :return: Dictionary of data
    """
    p = {BaseParticle(k) if isinstance(k, str) else k: v for k, v in p.items()}
    p = {i: p[i] for i in decs.outs}
    data_p = struct_momentum(p, center_mass=center_mass)
    if using_topology:
        decay_chain_struct = decs.topology_structure()
    else:
        decay_chain_struct = decs
    for dec in decay_chain_struct:
        data_p = infer_momentum(data_p, dec)
        # print(data_p)
        # exit()
        data_p = add_mass(data_p, dec)
    data_d = cal_angle_from_particle(
        data_p, decs, using_topology, r_boost=r_boost, random_z=random_z
    )
    data = {"particle": data_p, "decay": data_d}
    add_relative_momentum(data)
    return CalAngleData(data)


def prepare_data_from_dat_file4(fnames):
    """
    [deprecated] angle for amplitude4.py
    """
    a, b, c, d, e, f = [BaseParticle(i) for i in "ABCDEF"]
    bc, cd, bd = [BaseParticle(i) for i in ["BC", "CD", "BD"]]
    p = load_dat_file(fnames, [d, b, c, e, f])
    p = {i: p[i] for i in [b, c, e, f]}
    # st = {b: [b], c: [c], d: [d], a: [b, c, d], r: [b, d]}
    BaseDecay(a, [bc, d])
    BaseDecay(bc, [b, c])
    BaseDecay(a, [cd, b])
    BaseDecay(cd, [c, d])
    BaseDecay(a, [bd, c])
    BaseDecay(bd, [b, d])
    BaseDecay(d, [e, f])
    decs = DecayGroup(a.chain_decay())
    # decs = DecayChain.from_particles(a, [d, b, c])
    data = cal_angle_from_momentum(p, decs)
    data = data_to_numpy(data)
    data = flatten_dict_data(data)
    return data


def get_keys(dic, key_path=""):
    """get_keys of nested dictionary

    >>> a = {"a": 1, "b": {"c": 2}}
    >>> get_keys(a)
    ['/a', '/b/c']

    """
    keys_list = []

    def get_keys(dic, key_path):
        if isinstance(dic, dict):
            for i in dic:
                get_keys(dic[i], key_path + "/" + str(i))
        else:
            keys_list.append(key_path)

    get_keys(dic, key_path)
    return keys_list


def get_key_content(dic, key_path):
    """get key content. E.g. get_key_content(data, '/particle/(B, C)/m')

    >>> data = {"particle": {"(B, C)": {"p": 0.1, "m": 1.0},"B":1.0}}
    >>> get_key_content(data, '/particle/(B, C)/m')
    1.0

    """
    keys = key_path.strip("/").split("/")

    def get_content(dic, keys):
        if len(keys) == 0:
            return dic
        for k in dic:
            if str(k) == keys[0]:
                ret = get_content(dic[k], keys[1:])
                break
        return ret

    return get_content(dic, keys)



import numpy as np
import tensorflow as tf

from tf_pwa.breit_wigner import BW
from tf_pwa.cal_angle import cal_angle_from_momentum
from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import data_index, data_mask
from tf_pwa.phasespace import PhaseSpaceGenerator


def gauss(x, sigma):
    a = tf.exp(-(x ** 2) / (2 * sigma ** 2))
    return a


def abs2(x):
    r = tf.math.real(x)
    i = tf.math.imag(x)
    return r * r + i * i


def resolution_bw(m, m0, g0, sigma, m_min, m_max):
    """simple breit wigner with 100 point convolution of gauss function"""
    N = 100
    delta_min = -5 * sigma
    delta_max = 5 * sigma
    delta_sigma = (delta_max - delta_min) / (N - 1)
    ret = tf.zeros_like(m)
    zeros = tf.zeros_like(m)
    weights = []
    amps = []
    for i in range(N):
        delta = delta_min + delta_sigma * i
        m_i = delta + m
        amp = abs2(BW(m_i, m0, g0) + 10)  # BW(m) + NR
        w = tf.cast(gauss(delta, sigma), m.dtype)
        # print(w)
        # cut for mass range
        w = tf.where((m_i > m_min) & (m_i < m_max), w, zeros)
        # print(m_min, m_max)
        amps.append(amp * w)
        weights.append(w)
    # print(tf.reduce_sum(amps, axis=0), tf.reduce_sum(weights, axis=0))
    return tf.reduce_sum(amps, axis=0) / tf.reduce_sum(weights, axis=0)


def simple_selection(data, weight):
    rnd = tf.random.uniform(weight.shape, dtype=weight.dtype)
    cut = rnd * tf.reduce_max(weight) * 1.2 < weight
    cut_data = data_mask(data, cut)
    return cut_data


def main():
    sigma = 0.005

    config = ConfigLoader("config.yml")
    decay = config.get_decay()
    m0 = decay.top.get_mass()
    m1, m2, m3 = [i.get_mass() for i in decay.outs]

    print("mass: ", m0, " -> ", m1, m2, m3)

    phsp = PhaseSpaceGenerator(m0, [m1, m2, m3])
    p1, p2, p3 = phsp.generate(100000)

    angle = cal_angle_from_momentum({"B": p1, "C": p2, "D": p3}, decay)
    amp = config.get_amplitude()

    m_idx = config.get_data_index("mass", "R_BC")
    m_BC = data_index(angle, m_idx)
    R_BC = decay.get_particle("R_BC1")
    m_R, g_R = R_BC.get_mass(), R_BC.get_width()

    # import matplotlib.pyplot as plt
    # x = np.linspace(1.3, 1.7, 1000)
    # amp = R_BC.get_amp({"m": x}).numpy()
    # plt.plot(x, np.abs(amp)**2)
    # plt.show()

    print("mass: ", m_R, "width: ", g_R)
    amp_s2 = resolution_bw(m_BC, m_R, g_R, sigma, m1 + m2, m0 - m3)

    print("|A|*R: ", amp_s2)
    cut_data = simple_selection(angle, amp_s2)

    ps = [data_index(cut_data, ("particle", i, "p")).numpy() for i in "BCD"]
    np.savetxt(
        "data/data_origin.dat", np.transpose(ps, (1, 0, 2)).reshape((-1, 4))
    )

    p1, p2, p3 = phsp.generate(100000)
    np.savetxt(
        "data/phsp.dat", np.transpose([p1, p2, p3], (1, 0, 2)).reshape((-1, 4))
    )

    p1, p2, p3 = phsp.generate(50000)
    np.savetxt(
        "data/phsp_plot.dat",
        np.transpose([p1, p2, p3], (1, 0, 2)).reshape((-1, 4)),
    )


if __name__ == "__main__":
    main()



import math

from tf_pwa.err_num import *


def test_add():
    a = NumberError(1.0, 0.3)
    b = NumberError(2.0, 0.4)
    c = a + b
    assert c.value == 3.0
    assert c.error == 0.5
    d = b - a
    assert d.value == 1.0
    assert d.error == 0.5
    e = -a
    assert e.value == -1.0
    assert e.error == 0.3
    f = a - 1.0
    d = a + 3.0


def test_mul():
    a = NumberError(3.0, 0.3)
    b = NumberError(2.0, 0.4)
    c = a * b
    assert c.value == 6.0
    assert c.error == math.sqrt(1.8)
    d = a / b
    assert d.value == 1.5
    assert d.error == math.sqrt(0.1125)
    e = b ** 3
    assert e.value == 8.0
    assert abs(e.error - math.sqrt(23.04)) < 1e-7
    f = b ** a
    g = 3.0 ** a


def test_exp():
    a = NumberError(3.0, 0.3)
    b = NumberError(2.0, 0.4)
    c = a.exp()
    d = b.log()
    e = a.apply(math.sqrt)
    g = a.apply(math.sin, math.cos)
    print(g)


def test_cal_err():
    a = NumberError(3.0, 0.3)
    b = NumberError(2.0, 0.3)
    f = lambda x: x * 3.0 + x ** 2 * 2.0 + 1.0
    c = cal_err(f, a)
    g = lambda x, y: x + y + (x - y) * (x + y)
    d = cal_err(g, a, b)
    e = cal_err(g, 3.0, b)
    h = cal_err(g, a, 2.0)
    assert d.value == e.value
    assert e.value == h.value



"""
Basic Amplitude Calculations.
A partial wave analysis process has following structure:

DecayGroup: addition (+)
    DecayChain: multiplication (x)
        Decay, Particle(Propagator)

"""
# pylint: disable=unused-wildcard-import,unused-import
from .base import *
from .core import *
from .flatte import ParticleFlatte
from .Kmatrix import KmatrixSingleChannelParticle
from .split_ls import ParticleBWRLS



import matplotlib.pyplot as plt

from tf_pwa.experimental import factor_system as fs

from .test_full import gen_toy, toy_config


def test_partial_amp(toy_config):
    amp = toy_config.get_amplitude()
    phsp = toy_config.get_data("phsp")[0]
    tw = amp(phsp)
    phsp = toy_config.generate_phsp(100000)
    tw = amp(phsp)
    id_, pw = fs.get_all_partial_amp(amp, phsp, ["g_ls"])
    mas = phsp.get_mass("(B, C)").numpy()
    plt.hist(mas, 100, weights=tw.numpy(), histtype="step")
    for name, i in zip(id_, pw):
        plt.hist(mas, 100, weights=i.numpy(), histtype="step", label=str(name))
    plt.yscale("log")
    plt.legend()
    plt.ylim((0.1, None))
    plt.savefig("factor_system.png")



import copy
import functools
import random

import yaml

from tf_pwa.amp import (
    DecayChain,
    DecayGroup,
    HelicityDecay,
    get_decay,
    get_decay_chain,
    get_particle,
    split_particle_type,
)

from .base_config import BaseConfig


def set_min_max(dic, name, name_min, name_max):
    if name not in dic and name_min in dic and name_max in dic:
        dic[name] = (
            random.random() * (dic[name_max] - dic[name_min]) + dic[name_min]
        )


def decay_chain_cut_ls(decay):
    for i in decay:
        if isinstance(i, HelicityDecay):
            if len(i.get_ls_list()) == 0:
                return False, f"{i} ls not aviable {i.get_ls_list()}"
    return True, ""


def decay_chain_cut_mass(decay):
    for i in decay:
        if isinstance(i, HelicityDecay):
            if i.core.mass is None or any([j.mass is None for j in i.outs]):
                continue
            # print(i, i.core.mass, [j.mass for j in i.outs])
            if i.core.mass < sum([j.mass for j in i.outs]):
                return (
                    False,
                    f"{i} mass break {i.core.mass} < {[j.mass for j in i.outs]}",
                )
    return True, ""


class DecayConfig(BaseConfig):
    decay_chain_cut_list = {
        "ls_cut": decay_chain_cut_ls,
        "mass_cut": decay_chain_cut_mass,
    }

    def __init__(self, dic, share_dict={}):
        self.config = copy.deepcopy(dic)
        self.decay_chain_config = dic.get("decay_chain", {})
        self.share_dict = share_dict
        self.particle_key_map = {
            "Par": "P",
            "m0": "mass",
            "g0": "width",
            "J": "J",
            "P": "P",
            "spins": "spins",
            "bw": "model",
            "model": "model",
            "bw_l": "bw_l",
            "running_width": "running_width",
        }
        self.cut_list = self.config["data"].get("decay_chain_cut", ["ls_cut"])
        self.decay_key_map = {"model": "model"}
        self.dec = self.decay_item(self.config["decay"])
        (
            self.particle_map,
            self.particle_property,
            self.top,
            self.finals,
        ) = self.particle_item(self.config["particle"], share_dict)
        self.full_decay = DecayGroup(
            self.get_decay_struct(
                self.dec,
                self.particle_map,
                self.particle_property,
                self.top,
                self.finals,
                self.decay_chain_config,
            )
        )
        if self.config["data"].get("cp_trans", True):
            self.disable_allow_cc(self.full_decay)
        self.decay_struct = DecayGroup(
            self.get_decay_struct(self.dec, process_cut=False)
        )
        identical_particles = self.config["data"].get(
            "identical_particles", None
        )
        if identical_particles is not None:
            self.decay_struct.identical_particles = identical_particles
            self.full_decay.identical_particles = identical_particles

    @staticmethod
    def load_config(file_name, share_dict={}):
        if isinstance(file_name, dict):
            return copy.deepcopy(file_name)
        if isinstance(file_name, str):
            if file_name in share_dict:
                return DecayConfig.load_config(share_dict[file_name])
            with open(file_name) as f:
                ret = yaml.safe_load(f)
                if ret is None:
                    ret = {}
            return ret
        raise TypeError("not support config {}".format(type(file_name)))

    def get_decay(self, full=True):
        if full:
            return self.full_decay
        else:
            return self.decay_struct

    @staticmethod
    def _list2decay(core, outs):
        parts = []
        params = {}
        for j in outs:
            if isinstance(j, dict):
                for k, v in j.items():
                    params[k] = v
            else:
                parts.append(j)
        dec = {"core": core, "outs": parts, "params": params}
        return dec

    @staticmethod
    def decay_item(decay_dict):
        decs = []
        for core, outs in decay_dict.items():
            is_list = [isinstance(i, list) for i in outs]
            if all(is_list):
                for i in outs:
                    dec = DecayConfig._list2decay(core, i)
                    decs.append(dec)
            else:
                dec = DecayConfig._list2decay(core, outs)
                decs.append(dec)
        return decs

    @staticmethod
    def _do_include_dict(d, o, share_dict={}):
        s = DecayConfig.load_config(o, share_dict)
        for i in s:
            if i in d:
                if isinstance(d[i], dict):
                    s[i].update(d[i])
                    d[i] = s[i]
            else:
                d[i] = s[i]

    @staticmethod
    def particle_item_list(particle_list):
        particle_map = {}
        particle_property = {}
        for particle, candidate in particle_list.items():
            if isinstance(candidate, list):  # particle map
                if len(candidate) == 0:
                    particle_map[particle] = []
                for i in candidate:
                    if isinstance(i, str):
                        particle_map[particle] = particle_map.get(
                            particle, []
                        ) + [i]
                    elif isinstance(i, dict):
                        map_i, pro_i = DecayConfig.particle_item_list(i)
                        for k, v in map_i.items():
                            particle_map[k] = particle_map.get(k, []) + v
                        particle_property.update(pro_i)
                    else:
                        raise ValueError(
                            "value of particle map {} is {}".format(i, type(i))
                        )
            elif isinstance(candidate, dict):
                particle_property[particle] = candidate
            else:
                raise ValueError(
                    "value of particle {} is {}".format(
                        particle, type(candidate)
                    )
                )
        return particle_map, particle_property

    @staticmethod
    def particle_item(particle_list, share_dict={}):
        top = particle_list.pop("$top", None)
        finals = particle_list.pop("$finals", None)
        includes = particle_list.pop("$include", None)
        if includes:
            if isinstance(includes, list):
                for i in includes:
                    DecayConfig._do_include_dict(
                        particle_list, i, share_dict=share_dict
                    )
            elif isinstance(includes, str):
                DecayConfig._do_include_dict(
                    particle_list, includes, share_dict=share_dict
                )
            else:
                raise ValueError(
                    "$include must be string or list of string not {}".format(
                        type(includes)
                    )
                )
        particle_map, particle_property = DecayConfig.particle_item_list(
            particle_list
        )

        if isinstance(top, dict):
            particle_property.update(top)
        if isinstance(finals, dict):
            particle_property.update(finals)
        return particle_map, particle_property, top, finals

    def rename_params(self, params, is_particle=True):
        ret = {}
        if is_particle:
            key_map = self.particle_key_map
        else:
            key_map = self.decay_key_map
        for k, v in params.items():
            ret[key_map.get(k, k)] = v
        return ret

    def decay_chain_cut(self, decays):
        ret = []
        for i in decays:
            flag = True
            for name in self.cut_list:
                f = DecayConfig.decay_chain_cut_list[name]
                new_flag, msg = f(i)
                flag = flag and new_flag
                if not flag:
                    print(
                        "remove decay chain",
                        i,
                        "by",
                        name,
                        "\n\tbecause of",
                        msg,
                    )
                    break
            if flag:
                ret.append(i)
        return ret

    def get_decay_struct(
        self,
        decay,
        particle_map=None,
        particle_params=None,
        top=None,
        finals=None,
        chain_params={},
        process_cut=True,
    ):
        """  get decay structure for decay dict"""
        particle_map = particle_map if particle_map is not None else {}
        particle_params = (
            particle_params if particle_params is not None else {}
        )

        particle_set = {}

        def add_particle(name):
            if name in particle_set:
                return particle_set[name]
            params = particle_params.get(name, {})
            params = self.rename_params(params)
            set_min_max(params, "mass", "m_min", "m_max")
            set_min_max(params, "width", "g_min", "g_max")
            part = get_particle(name, **params)
            particle_set[name] = part
            return part

        def wrap_particle(name):
            name_list = particle_map.get(name, [name])
            return [add_particle(i) for i in name_list]

        def all_combine(out):
            if len(out) < 1:
                yield []
            else:
                for i in out[0]:
                    for j in all_combine(out[1:]):
                        yield [i] + j

        decs = []
        for dec in decay:
            core = wrap_particle(dec["core"])
            outs = [wrap_particle(j) for j in dec["outs"]]
            for i in core:
                for j in all_combine(outs):
                    dec_i = get_decay(i, j, **dec["params"])
                    decs.append(dec_i)

        top_tmp, finals_tmp = set(), set()
        if top is None or finals is None:
            top_tmp, res, finals_tmp = split_particle_type(decs)
        if top is None:
            top_tmp = list(top_tmp)
            assert len(top_tmp) == 1, "not only one top particle"
            top = list(top_tmp)[0]
        else:
            if isinstance(top, str):
                top = particle_set[top]
            elif isinstance(top, dict):
                keys = list(top.keys())
                assert len(keys) == 1
                top = particle_set[keys.pop()]
            else:
                return particle_set[str(top)]
        if finals is None:
            finals = list(finals_tmp)
        elif isinstance(finals, (list, dict)):
            finals = [particle_set[i] for i in finals]
        else:
            raise TypeError("{}: {}".format(finals, type(finals)))

        dec_chain = top.chain_decay()
        ret = []
        for i in dec_chain:
            if sorted(DecayChain(i).outs) == sorted(finals):
                all_params = chain_params.get("$all", {})
                dec_c = get_decay_chain(i, **all_params)
                ret.append(dec_c)
        if process_cut:
            return self.decay_chain_cut(ret)
        return ret

    def disable_allow_cc(self, decay_group):
        for decay_chain in decay_group:
            for decay in decay_chain:
                if hasattr(decay, "allow_cc"):
                    decay.allow_cc = False



import subprocess

support_info = [
    "memory.total",
    "memory.free",
    "memory.used",
    "name",  # 后三项还没
    "utilization.gpu",
    "utilization.memory",
]


def get_gpu_info(s):
    if s in support_info:
        cmd = "nvidia-smi --query-gpu={} --format=csv,noheader".format(s)
        ret = subprocess.getoutput(cmd)
        return ret.split("\n")
    else:
        raise Exception("Not support")


def get_gpu_total_memory(i=0):
    mem = get_gpu_info("memory.total")
    ret = mem[i].split(" ")[0]
    return float(ret)


def get_gpu_free_memory(i=0):
    mem = get_gpu_info("memory.free")
    ret = mem[i].split(" ")[0]
    return float(ret)


def get_gpu_used_memory(i=0):
    mem = get_gpu_info("memory.used")
    ret = mem[i].split(" ")[0]
    return float(ret)
